/**
 * perseus/connector: this is an implementation of some JCA-related technologies
 * (resource adapters and managers) for the ObjectWeb consortium.
 * Copyright (C) 2001-2002 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 */

package org.objectweb.perseus.connector.ra.fos;

import org.objectweb.perseus.connector.manager.standalone.SaManager;
import org.objectweb.perseus.fos.api.FosLoggerFactory;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Level;
import org.objectweb.util.monolog.api.Logger;

import javax.naming.NamingException;
import javax.naming.Reference;
import javax.resource.NotSupportedException;
import javax.resource.ResourceException;
import javax.resource.cci.Connection;
import javax.resource.cci.ConnectionFactory;
import javax.resource.cci.ConnectionSpec;
import javax.resource.cci.RecordFactory;
import javax.resource.cci.ResourceAdapterMetaData;
import javax.resource.spi.ConnectionManager;
import javax.resource.spi.ConnectionRequestInfo;
import javax.resource.spi.ManagedConnectionFactory;
import javax.security.auth.Subject;

/**
 * @author	S. Chassande-Barrioz, P. Dechamboux
 */
public class FosConnectionFactory
	implements ConnectionFactory, ResourceAdapterMetaData {
	/**
	 * The logger into which traces about FosConnectionFactory are produced.
	 */
	private Logger logger;
	/**
	 * The logger into which traces about FosConnectionImpl are produced.
	 */
	private Logger entityLogger;
	/**
	 * The FOS Adapter Version.
	 */
	private final static String FOSADAPTERVERSION = "1.01";
	/**
	 * The FOS Adapter Vendor Name (i.e. ObjectWeb !!).
	 */
	private final static String FOSADAPTERVENDORNAME = "The ObjectWeb Consortium";
	/**
	 * The FOS Adapter Name.
	 */
	private final static String FOSADAPTERNAME = "File Object Store (FOS) Adapter";
	/**
	 * A short description of the FOS Adapter.
	 */
	private final static String FOSADAPTERSHORTDESCRIPTION = "It provides a storage subsystem for storing data objects into files.\nThe approach assumes that each data object is stored within a different file.\nIt supports transactions (atomic I/Os through logging/recovery) as well as XA.";
	/**
	 * The FOS Spec Version (none yet).
	 */
	private final static String FOSSPECVERSION = "1.01";

	/**
	 * The JNDI Reference associated to this adapter.
	 */
	private Reference jndiReference = null;
	/**
	 * The ConnectionManager used to actually get Connection.
	 */
	private ConnectionManager connectionManager = null;
	/**
	 * The ManagedConnectionFactory associated to this ConnectionFactory.
	 */
	private FosManagedConnectionFactory managedConnectionFactory;

	/**
	 * Constructs a FosConnectionFactory. Associates a ConnectionManager with
	 * it, or creates a default one (if null).
	 * @param fmcf	The ManagedConnectionFactory to be associated with this
	 * 				ConnectionFactory.
	 * @param cm	The ConnectionManager to be associated with this
	 * 				ConnectionFactory. It may be null.
	 */
	FosConnectionFactory(FosManagedConnectionFactory fmcf, ConnectionManager cm)
		throws ResourceException {
		logger = fmcf.getLogger(FosLoggerFactory.CONNECTION, true);
		entityLogger = fmcf.getLogger(FosLoggerFactory.CONNECTION, false);
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG,
					   "Constructs a new FosConnectionFactory.");
		if (cm == null) {
			if (FosLoggerFactory.DEBUG)
				logger.log(BasicLevel.DEBUG,
					"Constructs a default ConnectionManager.");
			connectionManager = new SaManager();
			((SaManager) connectionManager).setManagedConnectionFactory(fmcf);
		} else {
			connectionManager = cm;
		}
		managedConnectionFactory = fmcf;
	}

	/**
	 * Creates a new FosConnectionImpl.
	 * Should add a Pool in the future ??
	 */
	FosConnectionImpl createConnection() throws ResourceException {
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG, "Creates a new FosConnectionImpl.");
		return new FosConnectionImpl(entityLogger, this);
	}

	// IMPLEMENTATION OF METHODS FROM THE (cci)ConnectionFactory INTERFACE

	/**
	 * Connection allocation is delegated to the ConnectionManager assigned to
	 * this factory.
	 */
	public Connection getConnection() throws ResourceException {
		return (Connection) connectionManager.allocateConnection(
			managedConnectionFactory,
			(ConnectionRequestInfo) null);
	}

	/**
	 * Connection allocation is delegated to the ConnectionManager assigned to
	 * this factory.
	 * @param spec	The connection is not taken into account within FOS.
	 */
	public Connection getConnection(ConnectionSpec spec) throws ResourceException {
		return (Connection) connectionManager.allocateConnection(
			managedConnectionFactory,
			(ConnectionRequestInfo) null);
	}

	/**
	 * Assigns a JNDI Reference to the adapter.
	 * @param reference	The JNDI Reference to be assigned.
	 */
	public void setReference(Reference reference) {
		jndiReference = reference;
	}

	/**
	 * Returns the JNDI Reference.
	 * @return	The JNDI Reference.
	 */
	public Reference getReference() throws NamingException {
		return jndiReference;
	}

	/**
	 * No support for record.
	 */
	public RecordFactory getRecordFactory() throws ResourceException {
		throw new NotSupportedException("FOS: no support for record.");
	}

	/**
	 * The FosConnectionFactory manages its metadata on its own.
	 */
	public ResourceAdapterMetaData getMetaData() throws ResourceException {
		return this;
	}

	// IMPLEMENTATION OF METHODS FROM THE (cci)ResourceAdapaterMetaData INTERFACE

	/**
	 * Gives access to the adapter version metadata for the FOS.
	 * @return	The FOS adapter version.
	 */
	public String getAdapterVersion() {
		return FOSADAPTERVERSION;
	}

	/**
	 * Gives access to the adapter vendor name metadata for the FOS.
	 * @return	The FOS adapter vendor name.
	 */
	public String getAdapterVendorName() {
		return FOSADAPTERVENDORNAME;
	}

	/**
	 * Gives access to the adapter name metadata for the FOS.
	 * @return	The FOS adapter name.
	 */
	public String getAdapterName() {
		return FOSADAPTERNAME;
	}

	/**
	 * Gives access to a short description of the FOS adapter.
	 * @return	The FOS short description.
	 */
	public String getAdapterShortDescription() {
		return FOSADAPTERSHORTDESCRIPTION;
	}

	/**
	 * Gives access to the specification version of the FOS.
	 * @return	The FOS specification version.
	 */
	public String getSpecVersion() {
		return FOSSPECVERSION;
	}

	/**
	 * No interaction is supported by the FOS adapter.
	 */
	public String[] getInteractionSpecsSupported() {
		return new String[0];
	}

	/**
	 * No support for record.
	 */
	public boolean supportsExecuteWithInputAndOutputRecord() {
		return false;
	}

	/**
	 * No support for record.
	 */
	public boolean supportsExecuteWithInputRecordOnly() {
		return false;
	}

	/**
	 * Yes, it does support local transaction demarcation.
	 */
	public boolean supportsLocalTransactionDemarcation() {
		return true;
	}
}