/**
 * perseus/connector: this is an implementation of some JCA-related technologies
 * (resource adapters and managers) for the ObjectWeb consortium.
 * Copyright (C) 2001-2002 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 */

package org.objectweb.perseus.connector.ra.fos;

import org.objectweb.perseus.fos.api.FosException;
import org.objectweb.perseus.fos.api.FosLoggerFactory;
import org.objectweb.perseus.fos.api.FosStructure;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import java.util.Iterator;
import javax.resource.ResourceException;
import javax.resource.cci.Connection;
import javax.resource.cci.ConnectionMetaData;
import javax.resource.cci.Interaction;
import javax.resource.cci.LocalTransaction;
import javax.resource.cci.ResultSetInfo;

/**
 * @author	S. Chassande-Barrioz, P. Dechamboux
 */
//public class FosConnectionImpl implements FileAccess, ConnectionMetaData {

public class FosConnectionImpl
	implements FosConnection, ConnectionMetaData {
	/**
	 * The logger into which traces about FosConnectionImpl are produced.
	 */
	private Logger logger;
	/**
	 * The name of this storage sub-system.
	 */
	private final static String EISPRODUCTNAME = "File Object Store (FOS)";
	/**
	 * The ConnectionFactory that have requested the allocation of this
	 * Connection.
	 */
	private FosConnectionFactory connectionFactory;
	/**
	 * The ManagedConnection associated with this Connection.
	 */
	private FosManagedConnection managedConnection = null;

	/**
	 * Constructs a FosConnectionImpl.
	 * @param el	The logger into which to produce Connection-related traces.
	 * @param fcf	The ConnectionFactory that has requested the Connection
	 * 				creation.
	 */
	FosConnectionImpl(Logger el, FosConnectionFactory fcf) {
		logger = el;
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG, "Constructs a new FosConnectionImpl.");
		connectionFactory = fcf;
	}

	/**
	 * Initializes a Connection to be linked with a ManagedConnection.
	 * @param mc	The ManagedConnection to be linked to.
	 */
	void initialize(FosManagedConnection mc) {
		managedConnection = mc;
	}

	// IMPLEMENTATION OF METHODS FROM THE (cci)Connection INTERFACE

	/**
	 * No support for Interaction.
	 */
	public Interaction createInteraction() throws ResourceException {
		throw new ResourceException("FOS: no support for Interaction.");
	}

	public LocalTransaction getLocalTransaction() throws ResourceException {
		if (managedConnection == null)
			throw new ResourceException("FOS: cannot get a LocalTransaction with no associated ManagedConnection.");
		return managedConnection;
	}

	/**
	 * The FosConnectionImpl manages the metadata on its own.
	 */
	public ConnectionMetaData getMetaData() throws ResourceException {
		return this;
	}

	/**
	 * No support for ResultSet.
	 */
	public ResultSetInfo getResultSetInfo() throws ResourceException {
		throw new ResourceException("FOS: no support for ResultSet.");
	}

	/**
	 * Closes this Connection. Dissociates from the ManagedConnection to which
	 * it is linked.
	 */
	public void close() throws ResourceException {
		if (managedConnection == null)
			throw new ResourceException("FOS: cannot close a Connection with no associated ManagedConnection.");
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG, "Closes Connection: " + this);
		managedConnection.dissociateConnection(this);
		managedConnection = null;
	}

	/**
	 * Sets the connection to the relevant auto-commit mode.
	 */
	public void setAutoCommit(boolean b) throws ResourceException {
		if (managedConnection == null)
			throw new ResourceException("FOS: no associated ManagedConnection.");
		managedConnection.setAutoCommit(b);
	}

	/**
	 * Gets the connection auto-commit mode.
	 */
	public boolean getAutoCommit() throws ResourceException {
		if (managedConnection == null)
			throw new ResourceException("FOS: no associated ManagedConnection.");
		return managedConnection.getAutoCommit();
	}

	// IMPLEMENTATION OF METHODS FROM THE (cci)ConnectionMetaData INTERFACE

	/**
	 * Returns the name of this storage sub-system.
	 * @return	The storage sub-system name.
	 */
	public String getEISProductName() throws ResourceException {
		return EISPRODUCTNAME;
	}

	/**
	 * This is the same version number as the adapter.
	 * @return	The storage sub-system version.
	 */
	public String getEISProductVersion() throws ResourceException {
		return connectionFactory.getAdapterVersion();
	}

	/**
	 * No support for user name.
	 * @return	The empty string.
	 */
	public String getUserName() throws ResourceException {
		return "";
	}

	// IMPLEMENTATION OF METHODS FROM THE FosAccess INTERFACE
	// Delegates execution to the associated ManagedConnection.

	/**
	 * Tests if the file associated to a persistent object exists. This file is
	 * specified by the directory under which it is stored under dbDir, and
	 * its name (or id) within this directory.
	 * @param dirof	The directory under dbDir where the file should be located.
	 * @param id	The name of the file to test (corresponding to the object
	 *				identifier).
	 * @return		true if the object file exists, else false.
	 */
	public boolean exist(String dirof, String id) throws FosException {
		if (managedConnection == null)
			throw new FosException("JCA managed level unreachable.",
				new ResourceException("FOS: no associated ManagedConnection."));
		return managedConnection.exist(dirof, id);
	}

	/**
	 * Tests if a directory that stores persistent objects exists.
	 * @param dirof	The directory under dbDir to test the existence.
	 * @return		true if the directory exists, else false.
	 */
	public boolean existDir(String dirof) throws FosException {
		if (managedConnection == null)
			throw new FosException("JCA managed level unreachable.",
				new ResourceException("FOS: no associated ManagedConnection."));
		return managedConnection.existDir(dirof);
	}

	/**
	 * Reads the content of a persistent object from a file. This file is
	 * specified by the directory under which it is stored under dbDir, and
	 * its name (or id) within this directory. The object that actually reads
	 * the file is also given by the user.
	 * @param dirof	The directory under dbDir where the read file is located.
	 * @param id	The name of the file to read (corresponding to the object
	 *				identifier).
	 * @param fs	The user object for actually reading the file.
	 */
	public void read(String dirof, String id, FosStructure fs, Object ctxt)
		throws FosException {
		if (managedConnection == null)
			throw new FosException("JCA managed level unreachable.",
				new ResourceException("FOS: no associated ManagedConnection."));
		managedConnection.read(dirof, id, fs, this, ctxt);
	}

	/**
	 * Deletes the file associated with a persistent object. This file is
	 * specified by the directory under which it is stored under dbDir, and
	 * its name (or id) within this directory.
	 * @param dirof	The directory under dbDir where the deleted file is
	 *				located.
	 * @param id	The name of the file to delete (corresponding to the object
	 *				identifier).
	 */
	public void delete(String dirof, String id) throws FosException {
		if (managedConnection == null)
			throw new FosException("JCA managed level unreachable.",
				new ResourceException("FOS: no associated ManagedConnection."));
		managedConnection.delete(dirof, id);
	}

	/**
	 * Deletes a directory that stores persistent objects along with all object
	 * files stored under it.
	 * @param dirof	The directory under dbDir to be deleted.
	 */
	public void deleteDir(String dirof) throws FosException {
		if (managedConnection == null)
			throw new FosException("JCA managed level unreachable.",
				new ResourceException("FOS: no associated ManagedConnection."));
		managedConnection.deleteDir(dirof);
	}

	/**
	 * Gets an iterator in order to iterate over the names of the data
	 * object files stored into that sub-directory.
	 * @param dirof	The sub-directory from which to scan the data
	 *		object file names.
	 * @return	The iterator in order to iterate over these names.
	 */
	public Iterator scan(String dirof) throws FosException {
		if (managedConnection == null)
			throw new FosException("JCA managed level unreachable.",
				new ResourceException("FOS: no associated ManagedConnection."));
		return managedConnection.scan(dirof);
	}

	/**
	 * Writes the content of a persistent object to a file. This file is
	 * specified by the directory under which it is stored under dbDir, and
	 * its name (or id) within this directory. The object that actually writes
	 * the file is also given by the user.
	 * @param dirof	The directory under dbDir where the written file is
	 *				located.
	 * @param id	The name of the file to write (corresponding to the object
	 *				identifier).
	 * @param fs	The user object for actually writing the file.
	 */
	public void write(String dirof, String id, FosStructure fs, Object ctxt)
		throws FosException {
		if (managedConnection == null)
			throw new FosException("JCA managed level unreachable.",
				new ResourceException("FOS: no associated ManagedConnection."));
		managedConnection.write(dirof, id, fs, this, ctxt);
	}
}