/**
 * perseus/connector: this is an implementation of some JCA-related technologies
 * (resource adapters and managers) for the ObjectWeb consortium.
 * Copyright (C) 2001-2002 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 */

package org.objectweb.perseus.connector.ra.fos;

import org.objectweb.perseus.fos.api.FosAccess;
import org.objectweb.perseus.fos.api.FosException;
import org.objectweb.perseus.fos.api.FosLoggerFactory;
import org.objectweb.perseus.fos.api.FosStructure;
import org.objectweb.perseus.fos.api.FosTransaction;
import org.objectweb.perseus.fos.lib.FosTxContext;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Iterator;
import javax.resource.ResourceException;
import javax.resource.spi.ConnectionEvent;
import javax.resource.spi.ConnectionEventListener;
import javax.resource.spi.ConnectionRequestInfo;
import javax.resource.spi.LocalTransaction;
import javax.resource.spi.ManagedConnection;
import javax.resource.spi.ManagedConnectionMetaData;
import javax.security.auth.Subject;
import javax.transaction.xa.XAResource;

/**
 * @author	S. Chassande-Barrioz, P. Dechamboux
 */
public class FosManagedConnection
	implements ManagedConnection, LocalTransaction,
	javax.resource.cci.LocalTransaction {
	private static final ArrayList NOLISTENERS = new ArrayList();
	/**
	 * The logger into which traces about FosManagedConnection are produced.
	 */
	private Logger logger;
	/**
	 * The ManagedConnectionFactory that have requested the allocation of this
	 * ManagedConnection.
	 */
	private FosManagedConnectionFactory managedConnectionFactory;
	/**
	 * The Connections associated with this ManagedConnection. Always at least
	 * one connection.
	 */
	private ArrayList associatedConnections = new ArrayList(1);
	/**
	 * The ConnectionEventListener list associated with this ManagedConnection.
	 */
	private ArrayList associatedListeners;
	/**
	 * It is set if the ManagedConnection execute within a LocalTransaction.
	 */
	private FosTxContext localTransactionTxContext;
	/**
	 * It is set if the ManagedConnection has been requested for a XAResource.
	 */
	private FosXAResource xaResource;
	/**
	 * It defines the auto-commit mode of this connection. It defaults to true
	 * at creation. It is ignored while running within a LocalTransaction or
	 * a DTP context (XA mode).
	 */
	private boolean autoCommit;

	/**
	 * Constructs a FosManagedConnection.
	 * @param el	The logger into which to produce ManagedConnection-related
	 * 				traces.
	 * @param fmcf	The ManagedConnectionFactory that has requested the
	 * 				ManagedConnection creation.
	 */
	FosManagedConnection(Logger el, FosManagedConnectionFactory fmcf) {
		logger = el;
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG,
					   "Constructs a new FosManagedConnection");
		managedConnectionFactory = fmcf;
		associatedListeners = NOLISTENERS;
		try {
			cleanup();
		} catch (ResourceException re) {
			// Can never happen
		}
	}

	/**
	 * Dissociates a Connection from the ones that are associated to this
	 * ManagedConnection.
	 * @param conn	The Connection to be dissociated.
	 */
	synchronized void dissociateConnection(Object conn) throws ResourceException {
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG, "Dissociates Connection: " + conn);
		int ind = associatedConnections.indexOf(conn);
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG, "Index within the associatedConnections: " + ind);
		if (ind == -1)
			throw new ResourceException("FOS: cannot dissociate an unknown Connection.");
		ConnectionEvent ce = new ConnectionEvent(this,
			ConnectionEvent.CONNECTION_CLOSED);
		ce.setConnectionHandle(conn);
		for (Iterator it = associatedListeners.iterator(); it.hasNext();)
			((ConnectionEventListener) it.next()).connectionClosed(ce);
		associatedConnections.remove(ind);
		if (associatedConnections.size() == 0)
			cleanup();
	}

	/**
	 * Sets the connection to the relevant auto-commit mode.
	 */
	public synchronized void setAutoCommit(boolean b) throws ResourceException {
		autoCommit = b;
	}

	/**
	 * Gets the connection auto-commit mode.
	 */
	public boolean getAutoCommit() throws ResourceException {
		return autoCommit;
	}

	// IMPLEMENTATION OF METHODS FROM THE (cci)ManagedConnection INTERFACE

	/**
	 * Delegates the creation of a Connection to the ConnectionFactory.
	 * "subject" and "info" parameters are ignored.
	 */
	public Object getConnection(Subject subject, ConnectionRequestInfo info)
		throws ResourceException {
		FosConnectionImpl rac = ((FosConnectionFactory) managedConnectionFactory.getConnectionFactory()).createConnection();
		rac.initialize(this);
		return rac;
	}

	/**
	 * Cleans up the connection.
	 */
	public void cleanup() throws ResourceException {
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG, "Cleans up ManagedConnection");
		int clsize = associatedConnections.size();
		if (clsize != 0) {
			while (clsize-- > 0)
				associatedConnections.remove(0);
			logger.log(BasicLevel.WARN, "Fos ManagedConnection still has associated Fos Connections (" + clsize + ")!");
		}
		if (associatedListeners != NOLISTENERS) {
			clsize = associatedListeners.size();
			if (clsize != 0) {
				while (clsize-- > 0)
					associatedListeners.remove(0);
				logger.log(BasicLevel.WARN, "Fos ManagedConnection still has associated Listeners (" + clsize + ")!");
			}
		}
		localTransactionTxContext = null;
		if (xaResource != null) {
			managedConnectionFactory.getXAResourceFactory().releaseXAResource(xaResource);
			xaResource = null;
		}
		autoCommit = true;
	}

	/**
	 * Called when ManagedConnection is removed.
	 */
	public void destroy() throws ResourceException {
		cleanup();
		logger = null;
		managedConnectionFactory = null;
		associatedConnections = null;
		associatedListeners = null;
		localTransactionTxContext = null;
		xaResource = null;
	}

	/**
	 * Associates a new Connection to this ManagedConnection. Nothing is done
	 * if it has already been associated (enforces no duplicate).
	 */
	public synchronized void associateConnection(Object o) throws ResourceException {
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG, "Associates Connection: " + o);
		if (!associatedConnections.contains(o))
			associatedConnections.add(o);
		else
			logger.log(BasicLevel.WARN, "Connection: " + o
				+ "has already been associated: ignored!");
	}

	/**
	 * Adds a listener to the associatedListeners list if it has not
	 * already been done. This means that duplicates are ignored.
	 * It first creates this list if needed (creation at first add).
	 */
	public synchronized void addConnectionEventListener(
		ConnectionEventListener listener) {
		if (associatedListeners == NOLISTENERS)
			associatedListeners = new ArrayList();
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG, "Associates ConnectionEventListener: "
				+ listener);
		if (!associatedListeners.contains(listener))
			associatedListeners.add(listener);
		else
			logger.log(BasicLevel.WARN, "ConnectionEventListener: " + listener
				+ "has already been associated: ignored!");
	}

	/**
	 * Removes a listener from the associatedListeners list. If the list does
	 * not eist (null), or if this listener does not belong to this list,
	 * nothing is done.
	 */
	public synchronized void removeConnectionEventListener(
		ConnectionEventListener listener) {
		if (associatedListeners == NOLISTENERS)
			logger.log(BasicLevel.WARN,
				"No associated ConnectionEventListener: ignored!");
		int ind = associatedListeners.indexOf(listener);
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG,
				"Index within the associatedListeners: " + ind);
		if (ind == -1)
			logger.log(BasicLevel.WARN, "ConnectionEventListener" + listener
				+ "not associated here: ignored!");
		else
			associatedListeners.remove(ind);
	}

	/**
	 * Retrieves an XA resource. Switches running mode to XA mode.
	 */
	public synchronized XAResource getXAResource() throws ResourceException {
		try {
			if (localTransactionTxContext != null)
				throw new ResourceException(
					"Try switching to XA running mode while a LocalTransaction has been begun!");
			if (xaResource == null) {
				xaResource = managedConnectionFactory.getXAResourceFactory().createXAResource();
				if (FosLoggerFactory.DEBUG)
					logger.log(BasicLevel.DEBUG, "Creates a new XAResource: " + xaResource);
			}
			return xaResource;
		} catch (FosException fe) {
			ResourceException re = new ResourceException("FOS: [nested exception].");
			re.setLinkedException(fe);
			throw re;
		}
	}

	public LocalTransaction getLocalTransaction() throws ResourceException {
		return this;
	}

	public ManagedConnectionMetaData getMetaData() throws ResourceException {
		return null;
	}

	/**
	 * Not supported yet.
	 */
	public void setLogWriter(PrintWriter writer) throws ResourceException {
		throw new ResourceException("FOS: logging to PrintWriter not supported yet.");
	}

	/**
	 * Not supported yet.
	 */
	public PrintWriter getLogWriter() throws ResourceException {
		throw new ResourceException("FOS: logging to PrintWriter not supported yet.");
	}

	// IMPLEMENTATION OF METHODS FROM THE (cci)LocalTransaction INTERFACE

	/**
	 * Begins the LocalTransaction if it has not already been started.
	 */
	public synchronized void begin() throws ResourceException {
		if (localTransactionTxContext != null)
			throw new ResourceException("FOS: a LocalTransaction has already begun.");
		try {
			localTransactionTxContext
				= (FosTxContext) managedConnectionFactory.getTxContextFactory()
				.createTxContext();
			localTransactionTxContext.begin();
			if (FosLoggerFactory.DEBUG)
				logger.log(BasicLevel.DEBUG, "LocalTransaction begin - txContext: "
					+ localTransactionTxContext);
			if (xaResource != null)
				logger.log(BasicLevel.WARN,
					"Begins a LocalTransaction on a ManagedConnection that runs in XA mode!");
			ConnectionEvent ce = new ConnectionEvent(this,
				ConnectionEvent.LOCAL_TRANSACTION_STARTED);
			for (Iterator it = associatedListeners.iterator(); it.hasNext();)
				((ConnectionEventListener) it.next()).localTransactionStarted(ce);
		} catch (FosException fe) {
			ResourceException re = new ResourceException(
				"FOS: cannot begin LocalTransaction [nested exception].");
			re.setLinkedException(fe);
			throw re;
		}
	}

	/**
	 * Commits the LocalTransaction if it is active.
	 */
	public synchronized void commit() throws ResourceException {
		if (localTransactionTxContext == null)
			throw new ResourceException("FOS: no LocalTransaction has been begun.");
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG, "LocalTransaction commit - txContext: "
				+ localTransactionTxContext);
		try {
			localTransactionTxContext.commit();
			managedConnectionFactory.getTxContextFactory().releaseTxContext(localTransactionTxContext);
			localTransactionTxContext = null;
			ConnectionEvent ce = new ConnectionEvent(this,
				ConnectionEvent.LOCAL_TRANSACTION_COMMITTED);
			for (Iterator it = associatedListeners.iterator(); it.hasNext();)
				((ConnectionEventListener) it.next()).localTransactionCommitted(ce);
		} catch (FosException fe) {
			ResourceException re = new ResourceException(
				"FOS: cannot commit LocalTransaction [nested exception].");
			re.setLinkedException(fe);
			throw re;
		}
	}

	/**
	 * Rollbacks the LocalTransaction if it is active.
	 */
	public synchronized void rollback() throws ResourceException {
		if (localTransactionTxContext == null)
			throw new ResourceException("FOS: no LocalTransaction has been begun.");
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG, "LocalTransaction rollback - txContext: "
				+ localTransactionTxContext);
		try {
			localTransactionTxContext.rollback();
			managedConnectionFactory.getTxContextFactory().releaseTxContext(localTransactionTxContext);
			localTransactionTxContext = null;
			ConnectionEvent ce = new ConnectionEvent(this,
				ConnectionEvent.LOCAL_TRANSACTION_ROLLEDBACK);
			for (Iterator it = associatedListeners.iterator(); it.hasNext();)
				((ConnectionEventListener) it.next()).localTransactionRolledback(ce);
		} catch (FosException fe) {
			ResourceException re = new ResourceException(
				"FOS: cannot rollback LocalTransaction [nested exception].");
			re.setLinkedException(fe);
			throw re;
		}
	}

	// IMPLEMENTATION OF METHODS FROM THE FosAccess INTERFACE

	/**
	 * Deletes the file associated with a persistent object. This file is
	 * specified by the directory under which it is stored under dbDir, and
	 * its name (or id) within this directory.
	 * @param dirof	The directory under dbDir where the deleted file is
	 *				located.
	 * @param id	The name of the file to delete (corresponding to the object
	 *				identifier).
	 */
	public synchronized void delete(String dirof, String id) throws FosException {
		if (localTransactionTxContext != null) {
			if (FosLoggerFactory.DEBUG)
				logger.log(BasicLevel.DEBUG, "Performs DELETE under LocalTransaction mode.");
			localTransactionTxContext.delete(dirof, id);
			return;
		}
		if (xaResource != null) {
			if (FosLoggerFactory.DEBUG)
				logger.log(BasicLevel.DEBUG, "Performs DELETE under XA mode.");
			xaResource.getTxContext().delete(dirof, id);
			return;
		}
		if (!autoCommit)
			throw new FosException(
				"Auto-commit OFF: cannot perform FosAccess operation with no transaction context!");
		// Executes in auto-commit mode.
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG, "Performs DELETE under AUTO-COMMIT mode.");
		FosTransaction txContext
			= managedConnectionFactory.getTxContextFactory().createTxContext();
		txContext.begin();
		try {
			txContext.delete(dirof, id);
			txContext.commit();
			managedConnectionFactory.getTxContextFactory().releaseTxContext(txContext);
		} catch (FosException fe) {
			txContext.rollback();
			managedConnectionFactory.getTxContextFactory().releaseTxContext(txContext);
			throw new FosException("Problem while running in auto-commit mode.", fe);
		}
	}

	/**
	 * Deletes a directory that stores persistent objects along with all object
	 * files stored under it.
	 * @param dirof	The directory under dbDir to be deleted.
	 */
	public void deleteDir(String dirof) throws FosException {
		FosTransaction txContext
			= managedConnectionFactory.getTxContextFactory().createTxContext();
		try {
			txContext.deleteDir(dirof);
			managedConnectionFactory.getTxContextFactory().releaseTxContext(txContext);
		} catch (FosException fe) {
			managedConnectionFactory.getTxContextFactory().releaseTxContext(txContext);
			throw fe;
		}
	}

	/**
	 * Tests if the file associated to a persistent object exists. This file is
	 * specified by the directory under which it is stored under dbDir, and
	 * its name (or id) within this directory.
	 * @param dirof	The directory under dbDir where the file should be located.
	 * @param id	The name of the file to test (corresponding to the object
	 *				identifier).
	 * @return		true if the object file exists, else false.
	 */
	public synchronized boolean exist(String dirof, String id) throws FosException {
		if (localTransactionTxContext != null) {
			if (FosLoggerFactory.DEBUG)
				logger.log(BasicLevel.DEBUG, "Performs EXIST under LocalTransaction mode.");
			return localTransactionTxContext.exist(dirof, id);
		}
		if (xaResource != null) {
			if (FosLoggerFactory.DEBUG)
				logger.log(BasicLevel.DEBUG, "Performs EXIST under XA mode.");
			return xaResource.getTxContext().exist(dirof, id);
		}
		if (!autoCommit)
			throw new FosException(
				"Auto-commit OFF: cannot perform FosAccess operation with no transaction context!");
		// Executes in auto-commit mode.
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG, "Performs EXIST under AUTO-COMMIT mode.");
		FosTransaction txContext
			= managedConnectionFactory.getTxContextFactory().createTxContext();
		txContext.begin();
		try {
			boolean res = txContext.exist(dirof, id);
			txContext.commit();
			managedConnectionFactory.getTxContextFactory().releaseTxContext(txContext);
			return res;
		} catch (FosException fe) {
			txContext.rollback();
			managedConnectionFactory.getTxContextFactory().releaseTxContext(txContext);
			throw new FosException("Problem while running in auto-commit mode.", fe);
		}
	}

	/**
	 * Tests if a directory that stores persistent objects exists.
	 * @param dirof	The directory under dbDir to test the existence.
	 * @return		true if the directory exists, else false.
	 */
	public boolean existDir(String dirof) throws FosException {
		FosTransaction txContext
			= managedConnectionFactory.getTxContextFactory().createTxContext();
		try {
			boolean res = txContext.existDir(dirof);
			managedConnectionFactory.getTxContextFactory().releaseTxContext(txContext);
			return res;
		} catch (FosException fe) {
			managedConnectionFactory.getTxContextFactory().releaseTxContext(txContext);
			throw fe;
		}
	}

	/**
	 * Reads the content of a persistent object from a file. This file is
	 * specified by the directory under which it is stored under dbDir, and
	 * its name (or id) within this directory. The object that actually reads
	 * the file is also given by the user.
	 * @param dirof	The directory under dbDir where the read file is located.
	 * @param id	The name of the file to read (corresponding to the object
	 *				identifier).
	 * @param fs	The user object for actually reading the file.
	 */
	public synchronized void read(String dirof, String id, FosStructure fs,
								  FosAccess conn, Object ctxt)
		throws FosException {
		if (localTransactionTxContext != null) {
			if (FosLoggerFactory.DEBUG)
				logger.log(BasicLevel.DEBUG, "Performs READ under LocalTransaction mode.");
			localTransactionTxContext.read(dirof, id, fs, conn, ctxt);
			return;
		}
		if (xaResource != null) {
			if (FosLoggerFactory.DEBUG)
				logger.log(BasicLevel.DEBUG, "Performs READ under XA mode.");
			xaResource.getTxContext().read(dirof, id, fs, conn, ctxt);
			return;
		}
		if (!autoCommit)
			throw new FosException(
				"Auto-commit OFF: cannot perform FosAccess operation with no transaction context!");
		// Executes in auto-commit mode.
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG, "Performs READ under AUTO-COMMIT mode.");
		FosTxContext txContext
			= (FosTxContext) managedConnectionFactory.getTxContextFactory()
			.createTxContext();
		txContext.begin();
		try {
			txContext.read(dirof, id, fs, conn, ctxt);
			txContext.commit();
			managedConnectionFactory.getTxContextFactory().releaseTxContext(txContext);
		} catch (FosException fe) {
			txContext.rollback();
			managedConnectionFactory.getTxContextFactory().releaseTxContext(txContext);
			throw new FosException("Problem while running in auto-commit mode.", fe);
		}
	}

	/**
	 * Gets an iterator in order to iterate over the names of the data
	 * object files stored into that sub-directory.
	 * @param dirof	The sub-directory from which to scan the data
	 *		object file names.
	 * @return	The iterator in order to iterate over these names.
	 */
	public synchronized Iterator scan(String dirof) throws FosException {
		if (localTransactionTxContext != null) {
			if (FosLoggerFactory.DEBUG)
				logger.log(BasicLevel.DEBUG, "Performs SCAN under LocalTransaction mode.");
			return localTransactionTxContext.scan(dirof);
		}
		if (xaResource != null) {
			if (FosLoggerFactory.DEBUG)
				logger.log(BasicLevel.DEBUG, "Performs SCAN under XA mode.");
			return xaResource.getTxContext().scan(dirof);
		}
		if (!autoCommit)
			throw new FosException(
				"Auto-commit OFF: cannot perform FosAccess operation with no transaction context!");
		// Executes in auto-commit mode.
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG, "Performs SCAN under AUTO-COMMIT mode.");
		FosTransaction txContext
			= managedConnectionFactory.getTxContextFactory().createTxContext();
		txContext.begin();
		try {
			Iterator res = txContext.scan(dirof);
			txContext.commit();
			managedConnectionFactory.getTxContextFactory().releaseTxContext(txContext);
			return res;
		} catch (FosException fe) {
			txContext.rollback();
			managedConnectionFactory.getTxContextFactory().releaseTxContext(txContext);
			throw new FosException("Problem while running in auto-commit mode.", fe);
		}
	}

	/**
	 * Writes the content of a persistent object to a file. This file is
	 * specified by the directory under which it is stored under dbDir, and
	 * its name (or id) within this directory. The object that actually writes
	 * the file is also given by the user.
	 * @param dirof	The directory under dbDir where the written file is
	 *				located.
	 * @param id	The name of the file to write (corresponding to the object
	 *				identifier).
	 * @param fs	The user object for actually writing the file.
	 */
	public synchronized void write(String dirof, String id, FosStructure fs,
								   FosAccess conn, Object ctxt)
		throws FosException {
		if (localTransactionTxContext != null) {
			if (FosLoggerFactory.DEBUG)
				logger.log(BasicLevel.DEBUG, "Performs WRITE under LocalTransaction mode.");
			localTransactionTxContext.write(dirof, id, fs, conn, ctxt);
			return;
		}
		if (xaResource != null) {
			if (FosLoggerFactory.DEBUG)
				logger.log(BasicLevel.DEBUG, "Performs WRITE under XA mode.");
			xaResource.getTxContext().write(dirof, id, fs, conn, ctxt);
			return;
		}
		if (!autoCommit)
			throw new FosException(
				"Auto-commit OFF: cannot perform FosAccess operation with no transaction context!");
		// Executes in auto-commit mode.
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG, "Performs WRITE under AUTO-COMMIT mode.");
		FosTxContext txContext
			= (FosTxContext) managedConnectionFactory.getTxContextFactory()
			.createTxContext();
		txContext.begin();
		try {
			txContext.write(dirof, id, fs, conn, ctxt);
			txContext.commit();
			managedConnectionFactory.getTxContextFactory().releaseTxContext(txContext);
		} catch (FosException fe) {
			txContext.rollback();
			managedConnectionFactory.getTxContextFactory().releaseTxContext(txContext);
			throw new FosException("Problem while running in auto-commit mode.", fe);
		}
	}
}