/**
 * perseus/connector: this is an implementation of some JCA-related technologies
 * (resource adapters and managers) for the ObjectWeb consortium.
 * Copyright (C) 2001-2002 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 */

package org.objectweb.perseus.connector.ra.fos;

import org.objectweb.perseus.fos.api.FosException;
import org.objectweb.perseus.fos.api.FosLoggerFactory;
import org.objectweb.perseus.fos.api.FosManager;
import org.objectweb.perseus.fos.lib.FosTxContextFactory;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.MonologFactory;
import org.objectweb.util.monolog.api.Loggable;
import org.objectweb.util.monolog.wrapper.printwriter.LoggerImpl;
import org.objectweb.util.monolog.Monolog;

import java.io.PrintWriter;
import java.util.Set;
import javax.resource.ResourceException;
import javax.resource.cci.ConnectionFactory;
import javax.resource.spi.ConnectionManager;
import javax.resource.spi.ConnectionRequestInfo;
import javax.resource.spi.ManagedConnection;
import javax.resource.spi.ManagedConnectionFactory;
import javax.security.auth.Subject;

/**
 * @author	S. Chassande-Barrioz, P. Dechamboux
 */
public class FosManagedConnectionFactory
	implements ManagedConnectionFactory, FosLoggerFactory, FosAttributeControler {
	/**
	 * The logger into which traces about FosManagedConnectionFactory are
	 * produced.
	 */
	private Logger logger;
	/**
	 * The logger into which traces about FosManagedConnection are
	 * produced.
	 */
	private Logger entityLogger;
	/**
	 * It is assumed that only one ConnectionFactory is actually created by
	 * a ManagedConnectionFactory.
	 */
	private FosConnectionFactory connectionFactory;
	/**
	 * The factory for managing FOS XA resources.
	 */
	private FosXAResourceFactory xaResourceFactory = null;
	/**
	 * The URL associated with the FOS controled by this resource adapter.
	 */
	private String connectionURL = null;
	/**
	 * The status wrt initialization at next start the FOS controled by this resource adapter.
	 */
	private boolean initializeAtStartUp = false;
	/**
	 * The factory for managing FOS transaction contexts.
	 */
	private FosManager txContextFactory = null;
	/**
	 * The factory for managing FOS transaction contexts.
	 */
	private ConnectionManager connectionManager = null;

	private MonologFactory monologFactory = null;

	private PrintWriter printWriter = null;

	/**
	 * Starts this FosManagedConnectionFactory.
	 */
	public void start() throws ResourceException {
		//Logging initialisation
		if (monologFactory == null) {
			if (Monolog.monologFactory == Monolog.getDefaultMonologFactory()) {
				monologFactory = Monolog.initialize();
			} else {
				monologFactory = Monolog.monologFactory;
			}
		}
		if (logger == null) {
			logger = monologFactory.getLogger("org.objectweb.perseus.jdo.ra");
		}

		try {
			if (txContextFactory == null) {
				txContextFactory = new FosTxContextFactory();
				txContextFactory.setDbDir(connectionURL);
				txContextFactory.setMonologFactory(monologFactory);
				if (initializeAtStartUp) {
					initializeAtStartUp = false;
					txContextFactory.initialize();
				}
			}
			txContextFactory.start();
			logger = getLogger(FosLoggerFactory.MANAGEDCONNECTION, true);
			entityLogger = getLogger(FosLoggerFactory.MANAGEDCONNECTION, false);
			if (FosLoggerFactory.DEBUG)
				logger.log(BasicLevel.DEBUG,
					"Starts a FosManagedConnectionFactory - ConnectionManager: "
					+ connectionManager);
			connectionFactory = new FosConnectionFactory(this, connectionManager);
		} catch (FosException fe) {
			ResourceException re
				= new ResourceException("Cannot start the ManagedConnectionFactory.");
			re.setLinkedException(fe);
			throw re;
		}
	}

	/**
	 * Stops this FosManagedConnectionFactory.
	 */
	public void stop() throws ResourceException {
		try {
			txContextFactory.stop();
		} catch (FosException fe) {
			ResourceException re
				= new ResourceException("Cannot stop the ManagedConnectionFactory.");
			re.setLinkedException(fe);
			throw re;
		}
	}

	/**
	 * Delegates the creation of a Connection to the ConnectionFactory.
	 */
	public Object createConnection() throws ResourceException {
		return connectionFactory.createConnection();
	}

	/**
	 * Retrieves the FosTxContextFactory.
	 */
	FosManager getTxContextFactory() {
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG, "TxContextFactory is: " + txContextFactory);
		return txContextFactory;
	}

	/**
	 * Retrieves the ConnectionFactory.
	 */
	ConnectionFactory getConnectionFactory() {
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG, "ConnectionFactory is: " + connectionFactory);
		return connectionFactory;
	}

	/**
	 * Retrieves the XAResourceFactory. If null, creates it. This means that no
	 * object related to XAResource management is created if not executing in
	 * XA-aware runtime environment.
	 */
	FosXAResourceFactory getXAResourceFactory() {
		if (xaResourceFactory == null) {
			if (FosLoggerFactory.DEBUG)
				logger.log(BasicLevel.DEBUG, "First time retrieving the XAResourceFactory: creates it!");
			xaResourceFactory = new FosXAResourceFactory(this, txContextFactory);
		}
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG, "XAResourceFactory is: " + xaResourceFactory);
		return xaResourceFactory;
	}

	// IMPLEMENTATION OF METHODS FROM THE (cci)ManagedConnectionFactory INTERFACE

	/**
	 * Creates a FosConnectionFactory; yields the existing one if any.
	 * @param cm	The ConnectionManager to be used by the created
	 * 				ConnectionFactory (may be null).
	 */
	public Object createConnectionFactory(ConnectionManager cm)
		throws ResourceException {
		if (connectionFactory == null) {
			connectionFactory = new FosConnectionFactory(this, cm);
			if (FosLoggerFactory.DEBUG)
				logger.log(BasicLevel.DEBUG,
					"Creation of the ConnectionFactory.");
		} else {
			if (FosLoggerFactory.DEBUG)
				logger.log(BasicLevel.DEBUG,
					"The ConnectionFactory has already been created.");
		}
		return connectionFactory;
	}

	/**
	 * Creates a FosConnectionFactory; yields the existing one if any.
	 */
	public Object createConnectionFactory() throws ResourceException {
		return createConnectionFactory(null);
	}

	/**
	 * Creates a new FosManagedConnection.
	 */
	public ManagedConnection createManagedConnection(
		Subject subject,
		ConnectionRequestInfo info) throws ResourceException {
		return new FosManagedConnection(entityLogger, this);
	}

	/**
	 * No matching rules supported. Always yields the first element of the set
	 * if any.
	 */
	public ManagedConnection matchManagedConnections(
		Set set,
		Subject subject,
		ConnectionRequestInfo info) throws ResourceException {
		if (set.size() == 0)
			return null;
		return (ManagedConnection) set.iterator().next();
	}

	/**
	 * Not supported yet.
	 */
	public void setLogWriter(PrintWriter writer) throws ResourceException {
		if (logger == null) {
			if (writer instanceof Loggable) {
				logger = ((Loggable) writer).getLogger();
				monologFactory = (MonologFactory)
						((Loggable) writer).getLoggerFactory();
			} else {
				LoggerImpl li = new LoggerImpl(writer);
				logger = li;
				monologFactory = li;
			}
		}
		printWriter = writer;
	}

	/**
	 * Not supported yet.
	 */
	public PrintWriter getLogWriter() throws ResourceException {
		return printWriter;
	}

	// IMPLEMENTATION OF METHODS FROM THE FosLoggerFactory INTERFACE

	/**
	 * Gives access to the logger associated to a given kind of entities managed
	 * by this FOS adapter, which are further classified into factories or
	 * non-factories.
	 * @param entity	The kind of entity (see CONNECTION, MANAGEDCONNECTION,
	 * 					XARESOURCE and XARESOURCE constants defined above).
	 * @param factory	Specifies if the concerned entity is a factory.
	 * @return	The relevant logger.
	 */
	public Logger getLogger(byte entity, boolean factory) {
		return txContextFactory.getLogger(entity, factory);
	}

	public void setMonologFactory(MonologFactory mf) {
		monologFactory = mf;
		txContextFactory.setMonologFactory(mf);
	}
	// IMPLEMENTATION OF METHODS FROM THE FosAttributeControler INTERFACE

	/**
	 * Gives access to the URL associated with a FOS adapter.
	 */
	public String getConnectionURL() {
		return connectionURL;
	}

	/**
	 * Changes the URL associated with a FOS adapter.
	 */
	public void setConnectionURL(String url) {
		connectionURL = url;
	}

	/**
	 * Gives access to the status wrt initialization at next start a FOS adapter.
	 */
	public boolean getInitializeAtStartUp() {
		return initializeAtStartUp;
	}

	/**
	 * Changes the status wrt initialization at next start a FOS adapter.
	 */
	public void setInitializeAtStartUp(boolean init) {
		initializeAtStartUp = init;
	}

	public ConnectionManager getConnectionManager() {
		return connectionManager;
	}

	/**
	 * Changes the ConnectionManager associated with a FOS adapter.
	 */
	public void setConnectionManager(ConnectionManager cm) {
		connectionManager = cm;
	}
}