/**
 * perseus/connector: this is an implementation of some JCA-related technologies
 * (resource adapters and managers) for the ObjectWeb consortium.
 * Copyright (C) 2001-2002 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 */

package org.objectweb.perseus.connector.ra.fos;

import org.objectweb.perseus.fos.api.FosException;
import org.objectweb.perseus.fos.api.FosLoggerFactory;
import org.objectweb.perseus.fos.api.FosTransaction;
import org.objectweb.perseus.fos.lib.FosTxContext;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import java.io.IOException;
import javax.transaction.xa.XAException;
import javax.transaction.xa.XAResource;
import javax.transaction.xa.Xid;

/**
 * @author	S. Chassande-Barrioz, P. Dechamboux
 */
public class FosXAResource implements XAResource {
	/**
	 * The logger into which traces about FosXAResource are produced.
	 */
	private Logger logger;
	/**
	 * The XID of the transaction associated with this XAResource.
	 * txContext and xid work together: they should be both null
	 * or not at the same time.
	 */
	Xid xid = null;
	/**
	 * The FosTransaction associated to the "xid" transaction.
	 * txContext and xid work together: they should be both null
	 * or not at the same time.
	 */
	FosTxContext txContext = null;
	/**
	 * A flag that specifies that "end" is called while "isSameRM" has never
	 * replied true. If true, it means that the XAResource has been kept by
	 * JTA in order to managed this transaction in a distributed scope.
	 */
	boolean endWithNoSameRM = true;
	/**
	 * Signals if prepared has been done.
	 */
	boolean prepared = false;
	/**
	 * The factory that has requtested the creation of this XAResource.
	 */
	FosXAResourceFactory xaResourceFactory;

	/**
	 * Constructs a FosTxContext.
	 * @param el	The logger to be used to produced XAResource-related traces.
	 * @param ftcf	The factory that has requested the creation of this
	 * 				FosTxContext.
	 */
	FosXAResource(Logger el, FosXAResourceFactory fxarf) throws FosException {
		logger = el;
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG, "Constructs a new FosXAResource.");
		xaResourceFactory = fxarf;
	}

	/**
	 *
	 */
	FosTxContext getTxContext() throws FosException {
		return txContext;
	}

	/**
	 * Assigns the relevant TxContext to this XAResource.
	 */
	private void getTxContext(Xid xid, boolean create)
		throws XAException {
		if (this.xid != null) {
			if (this.xid != xid) {
				// We are inside a start/end section: should be the same xid !!!
				throw new XAException(XAException.XAER_PROTO);
			}
			logger.log(BasicLevel.WARN, "XAResource already inside DTP context: "
				+ xid
				+ " - ignored.");
			return;
		}
		if (txContext != null)
			throw new XAException(XAException.XAER_PROTO);
		// Try getting the FosTxContext associated to this xid
		txContext = xaResourceFactory.getTxContext(xid);
		if (txContext == null) {
			if (!create) {
				// No transaction exist with this XID and must not create one!!
				throw new XAException(XAException.XAER_PROTO);
			}
			try {
				// Creates a new XA transaction with this XID
				txContext = xaResourceFactory.createTxContext(xid);
			} catch (FosException fe) {
				throw new XAException(XAException.XAER_RMERR);
			}
		}
		this.xid = xid;
	}

	// IMPLEMENTATION OF METHODS FROM THE (jta)XAResource INTERFACE

	// Management of transaction enrolment: start, end, isSameRM

	/**
	 * Assigns an actual FOS transaction context to the XAResource
	 * within the give DTP context.
	 */
	public void start(Xid xid, int i) throws XAException {
		getTxContext(xid, true);
	}

	/**
	 * Releases this XAResource if it is no more used.
	 */
	public void end(Xid xid, int i) throws XAException {
		if (this.xid != xid)
			throw new XAException(XAException.XAER_PROTO);
		if (endWithNoSameRM) {
			// First XAResource to be used within this XID: assumes it has been
			// registered by JTA !!
		} else {
			this.xid = null;
			txContext = null;
			endWithNoSameRM = true;
			xaResourceFactory.releaseXAResource(this);
		}
	}

	/**
	 * Used by JTA in order to verify that it has not already registered a
	 * XAResource to manage this transaction context from this RM. If it is the
	 * case, both XAResource are warned of this fact (endWithNoSameRM = false).
	 * This means that these XAResource will be released at "end" time.
	 * @resource	The resource to be compared against this one wrt RM.
	 */
	public boolean isSameRM(XAResource resource) throws XAException {
		if (resource instanceof FosXAResource)
			return false;
		if (((FosXAResource) resource).xaResourceFactory != xaResourceFactory)
			return false;
		((FosXAResource) resource).endWithNoSameRM = false;
		endWithNoSameRM = false;
		return true;
	}

	// Management of transaction termination: prepare, commit, rollback

	/**
	 * Prepares the underlying FosTxContext (prepare phase of the 2PC).
	 */
	public int prepare(Xid xid) throws XAException {
		getTxContext(xid, false);
		try {
			if (txContext.prepare()) {
				prepared = true;
				return XAResource.XA_RDONLY;
			}
			prepared = true;
			return XAResource.XA_OK;
		} catch (FosException e) {
			throw new XAException(XAException.XAER_RMERR);
		}
	}

	/**
	 * Commits the underlying FosTxContext (commit phase of the 2PC).
	 */
	public void commit(Xid xid, boolean b) throws XAException {
		getTxContext(xid, false);
		try {
			if (!prepared)
				throw new XAException(XAException.XAER_PROTO);
			txContext.commit();
			xaResourceFactory.releaseTxContext(xid);
		} catch (FosException e) {
			throw new XAException(XAException.XAER_RMERR);
		} finally {
			prepared = false;
			this.xid = null;
			txContext = null;
		}
	}

	/**
	 * Roolbacks the underlying FosTxContext.
	 */
	public void rollback(Xid xid) throws XAException {
		getTxContext(xid, false);
		try {
			txContext.rollback();
			xaResourceFactory.releaseTxContext(xid);
		} catch (FosException e) {
			throw new XAException(XAException.XAER_RMERR);
		} finally {
			prepared = false;
			this.xid = null;
			txContext = null;
		}
	}

	// Management of recovery: forget, recover

	/**
	 *
	 */
	public void forget(Xid xid) throws XAException {
	}

	/**
	 * Gets the Xid of distributed transactions to be recovered from the FOS
	 * manager.
	 */
	public Xid[] recover(int i) throws XAException {
		return xaResourceFactory.getXidForRecovery();
	}

	// Management of transaction timeout: getTransactionTimeout,
	// setTransactionTimeout

	/**
	 *
	 */
	public int getTransactionTimeout() throws XAException {
		return 0;
	}

	/**
	 *
	 */
	public boolean setTransactionTimeout(int i) throws XAException {
		return false;
	}
}
