/**
 * perseus/connector: this is an implementation of some JCA-related technologies
 * (resource adapters and managers) for the ObjectWeb consortium.
 * Copyright (C) 2001-2002 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 */

package org.objectweb.perseus.connector.ra.fos;

import org.objectweb.perseus.fos.api.FosException;
import org.objectweb.perseus.fos.api.FosLoggerFactory;
import org.objectweb.perseus.fos.api.FosManager;
import org.objectweb.perseus.fos.api.FosTransaction;
import org.objectweb.perseus.fos.lib.FosTxContext;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import java.util.HashMap;
import javax.transaction.xa.Xid;

/**
 * @author	S. Chassande-Barrioz, P. Dechamboux
 */
public class FosXAResourceFactory {
	/**
	 * The logger into which traces about FosXAResourceFactory are produced.
	 */
	private Logger logger;
	/**
	 * The logger into which traces about FosXAResource are produced.
	 */
	private Logger entityLogger;
	/**
	 * The factory for creating new FosTransaction.
	 */
	private FosManager txContextFactory;
	/**
	 * The hashed structure that stores FosTransaction that have been associated
	 * with a particular XID.
	 */
	private HashMap txContexts = new HashMap();

	/**
	 * Delegates recovery to the FOS manager.
	 */
	Xid[] getXidForRecovery() {
		return txContextFactory.getXidForRecovery();
	}

	/**
	 * Constructs a FosXAResourceFactory.
	 * @param flf	The logger factory in order to allocate relevant loggers for
	 * 				XAResource-related entities.
	 */
	FosXAResourceFactory(FosLoggerFactory flf, FosManager fm) {
		logger = flf.getLogger(FosLoggerFactory.XARESOURCE, true);
		entityLogger = flf.getLogger(FosLoggerFactory.XARESOURCE, false);
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG,
					   "Constructs a new FosXAResourceFactory");
		txContextFactory = fm;
	}

	/**
	 * Creates a new XAResource for coordinating transaction in a potentially
	 * multi-RM environment.
	 */
	FosXAResource createXAResource() throws FosException {
		FosXAResource res = new FosXAResource(entityLogger, this);
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG, "Creates a XAResource: " + res);
		return res;
	}

	/**
	 * Releases a XAResource that is not used any more.
	 * @param xar	The released XAResource.
	 */
	void releaseXAResource(FosXAResource xar) {
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG, "Releases a XAResource: " + xar);
	}

	/**
	 * Looks for a FosTxContext associated with the particular transaction.
	 * @param xid	The DTP transaction identifier.
	 */
	FosTxContext getTxContext(Xid xid) {
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG, "Looks for the TxContext associated with XID: " + xid);
		FosTxContext txc = (FosTxContext) txContexts.get(xid);
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG, "Found: " + txc);
		return txc;
	}

	/**
	 * Creates a FosTxContext and associates it with the given DTP transaction.
	 * @param xid	The DTP transaction identifier.
	 */
	FosTxContext createTxContext(Xid xid) throws FosException {
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG, "Creates a FosTxContext to be associated with XID: " + xid);
		FosTxContext txc = (FosTxContext) txContextFactory.createTxContext();
		txc.begin(xid);
		txContexts.put(xid, txc);
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG, "New one is: " + txc);
		return txc;
	}

	/**
	 * Releases a DTP transaction context and its related resources.
	 * @param xid	The DTP transaction identifier.
	 */
	void releaseTxContext(Xid xid) throws FosException {
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG, "Dissociates the FosTxContext associated with XID: " + xid);
		FosTxContext txc = getTxContext(xid);
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG, "Releases FosTxContext: " + txc);
		if (txc == null) {
			logger.log(BasicLevel.WARN, "Try to release unknow XID context: " + xid);
			return;
		}
		txContexts.remove(xid);
		txContextFactory.releaseTxContext(txc);
	}
}
