/**
 * Copyright (C) 2000
 */

package org.objectweb.perseus.fos.api;

import java.util.Iterator;

/**
 * Defines the means to manipulate transactional files for storing data objects.
 * These data objects are stored into files (one file per data object). There
 * are all stored into a root directory "dbdir". They can be stored directly
 * under it or into a sub-directory "dirof" of it (at any depth).
 * @author	S. Chassande-Barrioz, P. Dchamboux
 */
public interface FosAccess {
	/**
	 * Deletes the file associated with a persistent object. This file is
	 * specified by the directory under which it is stored under dbDir, and
	 * its name (or id) within this directory.
	 * @param dirof	The directory under dbDir where the deleted file is
	 *				located.
	 * @param id	The name of the file to delete (corresponding to the object
	 *				identifier).
	 */
	void delete(String dirof, String id) throws FosException;

	/**
	 * Deletes a directory that stores persistent objects along with all object
	 * files stored under it.
	 * @param dirof	The directory under dbDir to be deleted.
	 */
	void deleteDir(String dirof) throws FosException;

	/**
	 * Tests if the file associated to a persistent object exists. This file is
	 * specified by the directory under which it is stored under dbDir, and
	 * its name (or id) within this directory.
	 * @param dirof	The directory under dbDir where the file should be located.
	 * @param id	The name of the file to test (corresponding to the object
	 *				identifier).
	 * @return		true if the object file exists, else false.
	 */
	boolean exist(String dirof, String id) throws FosException;

	/**
	 * Tests if a directory that stores persistent objects exists.
	 * @param dirof	The directory under dbDir to test the existence.
	 * @return		true if the directory exists, else false.
	 */
	boolean existDir(String dirof) throws FosException;

	/**
	 * Reads the content of a persistent object from a file. This file is
	 * specified by the directory under which it is stored under dbDir, and
	 * its name (or id) within this directory. The object that actually reads
	 * the file is also given by the user.
	 * @param dirof	The directory under dbDir where the read file is located.
	 * @param id	The name of the file to read (corresponding to the object
	 *				identifier).
	 * @param fs	The user object for actually reading the file.
	 */
	void read(String dirof, String id, FosStructure fs, Object ctxt)
		throws FosException;

	/**
	 * Gets an iterator in order to iterate over the names of the data
	 * object files stored into that sub-directory.
	 * @param dirof	The sub-directory from which to scan the data
	 *		object file names.
	 * @return	The iterator in order to iterate over these names.
	 */
	Iterator scan(String dirof) throws FosException;

	/**
	 * Writes the content of a persistent object to a file. This file is
	 * specified by the directory under which it is stored under dbDir, and
	 * its name (or id) within this directory. The object that actually writes
	 * the file is also given by the user.
	 * @param dirof	The directory under dbDir where the written file is
	 *				located.
	 * @param id	The name of the file to write (corresponding to the object
	 *				identifier).
	 * @param fs	The user object for actually writing the file.
	 */
	void write(String dirof, String id, FosStructure fs, Object ctxt)
		throws FosException;
}
