/**
 * Copyright (C) 2000
 */

package org.objectweb.perseus.fos.lib;

import org.objectweb.perseus.fos.api.FosException;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.ArrayList;
import java.util.Iterator;
import javax.transaction.xa.Xid;

/**
 * Defines the file that stores the log of a transaction. It mainly stores
 * the XID of the DTP transaction it is involved in (if any), or null. Then
 * it stores the name of all Object Files that have been modified by this
 * transaction.
 * @author	S. Chassande-Barrioz, P. Dchamboux
 */
public class FosLogFile {
	private static final String TERMINATOR = "/$/\n";
	private File log;
	private FileOutputStream logFos = null;
	private ObjectOutputStream logOos = null;
	private boolean recovered = false;
	private boolean prepared = false;
	private ArrayList ofnList;
	private Xid xid = null;

	/**
	 * Constructs a log file controler.
	 * @param fn	The name of the log file.
	 */
	FosLogFile(String fn) {
		log = new File(fn);
	}

	/**
	 *
	 */
	void writeXid(Xid xid) throws Exception {
		if (logFos != null)
			throw new FosException("Already started to write this log file.");
		logFos = new FileOutputStream(log.getPath());
		logOos = new ObjectOutputStream(logFos);
		logOos.writeObject(xid);
	}

	/**
	 *
	 */
	void writeObjectFile(String ofn) throws Exception {
		if (logFos == null)
			throw new FosException("Not started to write this log file.");
		if (ofn != null)
			logOos.writeUTF(ofn);
	}

	/**
	 *
	 */
	void writeEnd() throws Exception {
		if (logFos == null)
			throw new FosException("Not started to write this log file.");
		logOos.writeUTF(TERMINATOR);
		logOos.flush();
		logFos.flush();
		logFos.getFD().sync();
		logFos.close();
		logOos = null;
		logFos = null;
	}

	/**
	 *
	 */
	void completed() throws Exception {
		log.delete();
		log = null;
	}

	/**
	 *
	 */
	void recover() throws Exception {
		if (recovered)
			return;
		if (!log.exists())
			return;
		ofnList = new ArrayList();
		ObjectInputStream ois
			= new ObjectInputStream(new FileInputStream(log.getPath()));
		try {
			xid = (Xid) ois.readObject();
			String fn = ois.readUTF();
			while (!fn.equals(TERMINATOR)) {
				ofnList.add(fn);
				fn = ois.readUTF();
			}
			prepared = true;
			recovered = true;
		} catch (Exception e) {
			// log file not consistent: not prepared correctly
			recovered = true;
		} finally {
			ois.close();
		}
	}

	/**
	 *
	 */
	Xid getXid() throws FosException {
		if (!recovered)
			throw new FosException("Operation to be performed after recovery.");
		return xid;
	}

	/**
	 *
	 */
	boolean isPrepared() throws FosException {
		if (!recovered)
			throw new FosException("Operation to be performed after recovery.");
		return prepared;
	}

	/**
	 *
	 */
	Iterator iterateModifiedOf() throws FosException {
		if (!recovered)
			throw new FosException("Operation to be performed after recovery.");
		return ofnList.iterator();
	}
}
