/**
 * Copyright (C) 2000
 */

package org.objectweb.perseus.fos.lib;

import org.objectweb.perseus.fos.api.FosException;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;

/**
 * Defines the way to control atomic I/O of a Data Object represented
 * by a File. The atomicity mechanism is based on shadowing. Each Data
 * Object is written into a File named "id.ai" where "ai" stands for "after
 * image". Then, committing a Data Object merely consists in moving this
 * image into the regular Data Object File named "id".
 * Deleting an Object File consists in creating a File name "id.dl" where
 * "dl" stands for deleted. Then commtting the deletion consists in
 * removing both "id" and "id.dl" files in that order.
 * @author	S. Chassande-Barrioz, P. Dchamboux
 */
public class FosObjectFile {
	final static String DELETSUFFIX = ".dl";
	final static String WRITESUFFIX = ".ai";
	private File fOf;
	private File shadowOf = null;
	private File deleteOf = null;
	private FileOutputStream fosOf = null;
	private ObjectOutputStream oosOf = null;

	/**
	 * Constructs a Data Object File I/O controler.
	 * @param fn	The name of the Data Object File.
	 */
	FosObjectFile(String fn) {
		fOf = new File(fn);
	}

	/**
	 * Recovers Object File status wrt repository status.
	 */
	void recover() throws Exception {
		deleteOf = new File(fOf.getPath() + DELETSUFFIX);
		if (deleteOf.exists())
			return;
		deleteOf = null;
		shadowOf = new File(fOf.getPath() + WRITESUFFIX);
		if (shadowOf.exists())
			return;
		shadowOf = null;
	}

	/**
	 * Commits any modification performed on this Object File.
	 */
	void commit() throws Exception {
		if (deleteOf != null) {
			// Commits deletion.
			fOf.delete();
			deleteOf.delete();
			deleteOf = null;
			return;
		}
		if (shadowOf != null) {
			// Commits writing.
			if (fosOf != null)
				throw new FosException("Cannot commit while writing - Object File: "
					+ shadowOf.getPath());
			fOf.delete();
			if (!shadowOf.renameTo(fOf))
				throw new FosException("Cannot move after image into regular one.");
			shadowOf = null;
		}
	}

	/**
	 * Returns the name of the Object File if it has been modified.
	 */
	String isModified() {
		if ((deleteOf != null) || (shadowOf != null))
			return fOf.getPath();
		return null;
	}

	/**
	 * Rollbacks any modification performed on this Object File.
	 */
	synchronized void rollback() throws Exception {
		if (deleteOf != null) {
			// Rollbacks deletion.
			deleteOf.delete();
			deleteOf = null;
			return;
		}
		if (shadowOf != null) {
			// Rollbacks writing.
			if (fosOf != null) {
				fosOf.close();
				oosOf = null;
				fosOf = null;
			}
			shadowOf.delete();
			shadowOf = null;
		}
	}

	/**
	 * Opens a stream in order to read from this Object File. It should be
	 * closed at any time.
	 * @return	The stream to read from.
	 */
	ObjectInputStream openReader() throws Exception {
		return new ObjectInputStream(new FileInputStream(fOf.getPath()));
	}

	/**
	 * Opens a stream in order to write to this Object File. Ensures that
	 * there is only one writer at a time.
	 * @return	The stream to write to.
	 */
	synchronized ObjectOutputStream openWriter() throws Exception {
		if (deleteOf != null) {
			// There has been a delete operation before. Forgets deletion.
			deleteOf.delete();
			deleteOf = null;
		} else if (fosOf != null)
			throw new FosException("Cannot write in parallel - Object File: "
				+ shadowOf.getPath());
		if (shadowOf == null)
			shadowOf = new File(fOf.getPath() + WRITESUFFIX);
		fosOf = new FileOutputStream(shadowOf.getPath());
		return oosOf = new ObjectOutputStream(fosOf);
	}

	/**
	 * Closes the stream that has been open for writing. It must be done before
	 * any transaction-related action is performed.
	 */
	synchronized void closeWriter() throws Exception {
		oosOf.flush();
		fosOf.close();
		oosOf = null;
		fosOf = null;
	}

	/**
	 * Verifies if the Data Object File already exists.
	 */
	boolean exist() {
		if (deleteOf != null)
		// Already deleted.
			return false;
		if (shadowOf != null) {
			// There has been a write operation before. Forgets writing.
			return true;
		}
		return fOf.exists();
	}

	/**
	 * Deletes the Object File.
	 */
	synchronized void delete() throws Exception {
		if (deleteOf != null)
		// Already deleted.
			return;
		if (fosOf != null)
			throw new FosException("Cannot delete while writing - Object File: "
				+ shadowOf.getPath());
		if (shadowOf != null) {
			// There has been a write operation before. Forgets writing.
			shadowOf.delete();
			shadowOf = null;
		}
		deleteOf = new File(fOf.getPath() + DELETSUFFIX);
		deleteOf.createNewFile();
	}
}