/**
 * Copyright (C) 2000
 */

package org.objectweb.perseus.fos.lib;

import org.objectweb.perseus.fos.api.FosAccess;
import org.objectweb.perseus.fos.api.FosException;
import org.objectweb.perseus.fos.api.FosLoggerFactory;
import org.objectweb.perseus.fos.api.FosStructure;
import org.objectweb.perseus.fos.api.FosTransaction;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

import java.io.File;
import java.io.ObjectInputStream;
import java.io.IOException;
import java.util.HashMap;
import java.util.Iterator;
import javax.transaction.xa.Xid;

/**
 * @author	S. Chassande-Barrioz, P. D?chamboux
 */
public class FosTxContext implements FosTransaction {
	/**
	 * The logger into which traces about FosTxContext are produced.
	 */
	private Logger logger;
	/**
	 * The FosManager associated with this FosTransaction.
	 */
	private FosTxContextFactory txContextFactory;
	/**
	 * The Xid associated with this transaction. It is null when executing in
	 * non XA-aware environment.
	 */
	private Xid xid;
	/**
	 * The termination status of the FOS transaction. It is true after either
	 * commit or rollback has been executed.
	 */
	private boolean active;
	/**
	 * Signals if prepared has been done.
	 */
	private boolean prepared;
	/**
	 * Signals if any Object File has been written within this transaction.
	 */
	private boolean updated;
	/**
	 * The list of Object File controlers involved in this transaction.
	 */
	private HashMap fofList;
	/**
	 *
	 */
	private FosLogFile logFile;

	/**
	 * Constructs a FosTxContext.
	 * @param el	The logger to be used to produced TxContext-related traces.
	 * @param ftcf	The factory that has requested the creation of this
	 * 				FosTxContext.
	 */
	FosTxContext(Logger el, FosTxContextFactory ftcf) throws FosException {
		logger = el;
		txContextFactory = ftcf;
		xid = null;
		active = false;
		prepared = false;
		updated = false;
		fofList = null;
		logFile = null;
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG, "Constructs a new FosTxContext.");
	}

	/**
	 * Delivers the Object File associated with the given id within the given
	 * directory of DbDir.
	 */
	private synchronized FosObjectFile getOf(String ofn) {
		FosObjectFile fof = (FosObjectFile) fofList.get(ofn);
		if (fof == null) {
			fof = new FosObjectFile(ofn);
			fofList.put(ofn, fof);
		}
		return fof;
	}

	/**
	 * Specifies if this FOS transaction is active and has been prepared.
	 */
	boolean isPrepared() {
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG, "FOS Transaction "
				+ (prepared ? "prepared." : "not prepared."));
		return prepared;
	}

	/**
	 * Returns the DTP id associated with this FOS transaction.
	 */
	Xid getXid() {
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG, "FOS Transaction DTP id: " + xid);
		return xid;
	}

	/**
	 * Recreates the context for managing the recovery at cold restart.
	 * @return	The Xid if it is a DTP transaction for which no local decision
	 *			can be made wrt to its completion.
	 */
	Xid recover(String txid) throws FosException {
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG, "Recovers transaction: " + txid);
		logFile = new FosLogFile(txid);
		try {
			logFile.recover();
			xid = logFile.getXid();
			prepared = logFile.isPrepared();
			Iterator it = logFile.iterateModifiedOf();
			FosObjectFile fof;
			String ofn;
			while (it.hasNext()) {
				ofn = (String) it.next();
				if (FosLoggerFactory.DEBUG)
					logger.log(BasicLevel.DEBUG, "Recovers ObjectFile: " + ofn);
				fof = new FosObjectFile(ofn);
				fof.recover();
				if (!prepared)
					fof.rollback();
				else if (xid == null)
					fof.commit();
			}
			if (prepared && (xid != null)) {
				if (FosLoggerFactory.DEBUG)
					logger.log(BasicLevel.DEBUG, "DTP - cannot do local decision: " + xid);
				return xid;
			}
			logFile.completed();
			return null;
		} catch (Exception e) {
			throw new FosException("Cannot recover transaction [nested exception].", e);
		}
	}

	// IMPLEMENTATION OF METHODS FROM THE FosTransaction INTERFACE

	/**
	 * Begins a FOS transaction with no DTP context.
	 */
	public void begin() throws FosException {
		begin(null);
	}

	/**
	 * Begins a FOS transaction with a DTP context.
	 * @param xid	The DTP identifier associated with this FOS transaction. It
	 * 				is null if non XA runtime environment.
	 */
	public void begin(Xid xid) throws FosException {
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG, "Begins a FOS transaction - XID: " + xid);
		txContextFactory.setActiveTX(true);
		this.xid = xid;
		active = true;
		prepared = false;
		updated = false;
		fofList = new HashMap();
		if (xid != null) {
			try {
				logFile = new FosLogFile(txContextFactory.getTxId());
				logFile.writeXid(xid);
			} catch (Exception e) {
				throw new FosException("Cannot begin transaction [nested exception].", e);
			}
		} else
			logFile = null;
	}

	/**
	 * Specifies if this FOS transaction is active or not. true means that
	 * it has begun but it has not been committed or rollbacked yet.
	 */
	public boolean isActive() {
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG, "FOS Transaction "
				+ (active ? "active." : "inactive."));
		return active;
	}

	/**
	 * Prepares a FOS transaction to commit.
	 * @return	true if all objects used by this transaction was just read.
	 */
	public boolean prepare() throws FosException {
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG, "Prepares a FOS transaction - XID: " + xid);
		prepared = true;
		if (!updated) {
			if (FosLoggerFactory.DEBUG)
				logger.log(BasicLevel.DEBUG, "Read-Only transaction!");
			return true;
		}
		try {
			if (logFile == null) {
				logFile = new FosLogFile(txContextFactory.getTxId());
				logFile.writeXid(null);
			}
			Iterator it = fofList.values().iterator();
			while (it.hasNext())
				logFile.writeObjectFile(((FosObjectFile) it.next()).isModified());
			logFile.writeEnd();
			return false;
		} catch (Exception e) {
			throw new FosException("Cannot prepare transaction [nested exception].", e);
		}
	}

	/**
	 * Commits a FOS transaction.
	 */
	public void commit() throws FosException {
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG, "Commits a FOS transaction - XID: " + xid);
		if (!active)
			throw new FosException("Cannot commit inactive transaction.");
		if (!prepared)
			prepare();
		txContextFactory.setActiveTX(false);
		try {
			if (!updated) {
				if (FosLoggerFactory.DEBUG)
					logger.log(BasicLevel.DEBUG, "Read-Only transaction!");
			} else {
				logFile.completed();
				Iterator it = fofList.values().iterator();
				while (it.hasNext())
					((FosObjectFile) it.next()).commit();
			}
			xid = null;
			active = false;
			prepared = false;
			updated = false;
			fofList = null;
			logFile = null;
		} catch (Exception e) {
			throw new FosException("Cannot commit transaction [nested exception].", e);
		}
	}

	/**
	 * Roolbacks a FOS transaction.
	 */
	public void rollback() throws FosException {
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG, "Rollbacks a FOS transaction - XID: " + xid);
		if (!active)
			throw new FosException("Cannot rollback inactive transaction.");
		txContextFactory.setActiveTX(false);
		try {
			if (!updated) {
				if (FosLoggerFactory.DEBUG)
					logger.log(BasicLevel.DEBUG, "Read-Only transaction!");
			} else {
				if (logFile != null)
					logFile.completed();
				Iterator it = fofList.values().iterator();
				while (it.hasNext())
					((FosObjectFile) it.next()).rollback();
			}
			xid = null;
			active = false;
			prepared = false;
			updated = false;
			fofList = null;
			logFile = null;
		} catch (Exception e) {
			throw new FosException("Cannot rollback transaction [nested exception].", e);
		}
	}

	// IMPLEMENTATION OF METHODS FROM THE FosAccess INTERFACE

	/**
	 * Tests if the file associated to a persistent object exists. This file is
	 * specified by the directory under which it is stored under dbDir, and
	 * its name (or id) within this directory.
	 * @param dirof	The directory under dbDir where the file should be located.
	 * @param id	The name of the file to test (corresponding to the object
	 *				identifier).
	 * @return		true if the object file exists, else false.
	 */
	public boolean exist(String dirof, String id) throws FosException {
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG, "EXIST operation: dirof=\"" + dirof
				+ "\", id=\"" + id + "\"");
		if (prepared || !active)
			throw new FosException("FOS transaction context not ready to perform operations.");
		FosObjectFile fof = getOf(txContextFactory.getDbDir() + File.separator
			+ dirof + File.separator + id);
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG, "Object " + (fof.exist() ? "exists" : "does not exist"));
		return fof.exist();
	}

	/**
	 * Tests if a directory that stores persistent objects exists.
	 * @param dirof	The directory under dbDir to test the existence.
	 * @return		true if the directory exists, else false.
	 */
	public boolean existDir(String dirof) throws FosException {
		return (new File(txContextFactory.getDbDir() + File.separator + dirof)).exists();
	}

	/**
	 * Reads the content of a persistent object from a file. This file is
	 * specified by the directory under which it is stored under dbDir, and
	 * its name (or id) within this directory. The object that actually reads
	 * the file is also given by the user.
	 * @param dirof	The directory under dbDir where the read file is located.
	 * @param id	The name of the file to read (corresponding to the object
	 *				identifier).
	 * @param fs	The user object for actually reading the file.
	 */
	public void read(String dirof, String id, FosStructure fs, Object ctxt)
		throws FosException {
		read(dirof, id, fs, this, ctxt);
	}

	public void read(String dirof, String id, FosStructure fs,
					 FosAccess conn, Object ctxt) throws FosException {
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG, "READ operation: dirof=\"" + dirof
				+ "\", id=\"" + id + "\", FosStructure=" + fs);
		if (prepared || !active)
			throw new FosException("FOS transaction context not ready to perform operations.");
		if ((dirof == null) || (id == null) || (fs == null))
			throw new FosException("Wrong parameters.");
		FosObjectFile fof = getOf(txContextFactory.getDbDir() + File.separator
			+ dirof + File.separator + id);
        if (!fof.exist()) {
            throw new FosException("Cannot read: Data Object does not exist.");
        }
        ObjectInputStream ois = null;
		try {
			ois = fof.openReader();
			fs.readFile(ois, conn, ctxt);
			ois.close();
            ois = null;
		} catch (Exception e) {
            if (ois != null) {
                try {
                    ois.close();
                } catch (IOException e2) {
                }
            }
			throw new FosException("I/O problem [nested exception].", e);
		}
	}

	/**
	 * Deletes the file associated with a persistent object. This file is
	 * specified by the directory under which it is stored under dbDir, and
	 * its name (or id) within this directory.
	 * @param dirof	The directory under dbDir where the deleted file is
	 *				located.
	 * @param id	The name of the file to delete (corresponding to the object
	 *				identifier).
	 */
	public void delete(String dirof, String id) throws FosException {
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG, "REMOVE operation: dirof=\"" + dirof
				+ "\", id=\"" + id + "\"");
		if (prepared || !active)
			throw new FosException("FOS transaction context not ready to perform operations.");
		if ((dirof == null) || (id == null))
			throw new FosException("Wrong parameters.");
		FosObjectFile fof = getOf(txContextFactory.getDbDir() + File.separator
			+ dirof + File.separator + id);
		try {
			if (!fof.exist())
				throw new FosException("Cannot remove: Data Object does not exist.");
			if (FosLoggerFactory.DEBUG)
				logger.log(BasicLevel.DEBUG, "Really removes!!");
			fof.delete();
			updated = true;
		} catch (Exception e) {
			throw new FosException("I/O problem [nested exception].", e);
		}
	}

	/**
	 * Deletes a directory that stores persistent objects along with all object
	 * files stored under it.
	 * @param dirof	The directory under dbDir to be deleted.
	 */
	public void deleteDir(String dirof) throws FosException {
		String dirpath = txContextFactory.getDbDir() + File.separator + dirof;
		synchronized(txContextFactory) {
			if (txContextFactory.getNbActiveTX() != 0)
				throw new FosException("Cannot perform deleteDir: some TX are still active");
			if (! existDir(dirof))
				throw new FosException("Cannot perform deleteDir: directory does not exist");
			try {
				txContextFactory.deleteDirContent(new File(dirpath));
			} catch (Exception e) {
				throw new FosException("Fails performing deleteDir", e);
			}
		}
	}

	/**
	 * Gets an iterator in order to iterate over the names of the data
	 * object files stored into that sub-directory.
	 * @param dirof	The sub-directory from which to scan the data
	 *		object file names.
	 * @return	The iterator in order to iterate over these names.
	 */
	public Iterator scan(String dirof) throws FosException {
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG, "SCANEXTENSION operation: dirof=\""
				+ dirof + "\"");
		if (prepared || !active)
			throw new FosException("FOS transaction context not ready to perform operations.");
		if (dirof == null)
			throw new FosException("Wrong parameters.");
		return new FosExtension(txContextFactory.getDbDir() + File.separator
			+ dirof);
	}

	/**
	 * Writes the content of a persistent object to a file. This file is
	 * specified by the directory under which it is stored under dbDir, and
	 * its name (or id) within this directory. The object that actually writes
	 * the file is also given by the user.
	 * @param dirof	The directory under dbDir where the written file is
	 *				located.
	 * @param id	The name of the file to write (corresponding to the object
	 *				identifier).
	 * @param fs	The user object for actually writing the file.
	 */
	public void write(String dirof, String id, FosStructure fs, Object ctxt)
		throws FosException {
		write(dirof, id, fs, this, ctxt);
	}

	public void write(String dirof, String id, FosStructure fs,
					  FosAccess conn, Object ctxt) throws FosException {
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG, "WRITE operation: dirof=\"" + dirof
				+ "\", id=\"" + id + "\", FosStructure=" + fs);
		if (prepared || !active)
			throw new FosException("FOS transaction context not ready to perform operations.");
		if ((dirof == null) || (id == null) || (fs == null))
			throw new FosException("Wrong parameters.");
		File od = new File(txContextFactory.getDbDir() + File.separator + dirof);
		if (!od.exists()) {
			if (FosLoggerFactory.DEBUG)
				logger.log(BasicLevel.DEBUG, "Create ObjectFile DIR: " + od.getPath());
			od.mkdirs();
		}
		FosObjectFile fof = getOf(txContextFactory.getDbDir() + File.separator
			+ dirof + File.separator + id);
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG, "FosObjectFile: " + fof);
		try {
			fs.writeFile(fof.openWriter(), conn, ctxt);
			fof.closeWriter();
			updated = true;
		} catch (Exception e) {
			throw new FosException("I/O problem [nested exception].", e);
		}
	}
}
