/**
 * Copyright (C) 2000
 */

package org.objectweb.perseus.fos.lib;

import org.objectweb.perseus.fos.api.FosException;
import org.objectweb.perseus.fos.api.FosLoggerFactory;
import org.objectweb.perseus.fos.api.FosManager;
import org.objectweb.perseus.fos.api.FosTransaction;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.MonologFactory;

import java.io.File;
import java.util.ArrayList;
import javax.transaction.xa.Xid;

/**
 * @author	S. Chassande-Barrioz, P. D�chamboux
 */
public class FosTxContextFactory implements FosManager, FosLoggerFactory {
	/**
	 * The directory under dbDir into which log files are stored.
	 */
	private final static String TXDIR = ".fostx";
	/**
	 * For creating all loggers related to FOS and its adapter.
	 */
	private MonologFactory loggerFactory;
	/**
	 * The logger into which traces about FosTxContextFactory are produced.
	 */
	private Logger logger;
	/**
	 * The logger into which traces about FosTxContext are produced.
	 */
	private Logger entityLogger;
	/**
	 * The directory under which is stored all information related to the FOS.
	 */
	private String dbDir = null;
	/**
	 * The next transaction identifier to be delivered.
	 */
	private long txId;
	/**
	 * The transactions that have been created and not released.
	 */
	private ArrayList txList;
	/**
	 * The list of Xid of distributed transactions that need to be recovered.
	 */
	private ArrayList xidList = null;
	/**
	 *
	 */
	private int nbActiveTx = 0;

	/**
	 * Starts a FosTxContextFactory. It either initializes the FOS repository
	 * if it is the first start. In other cases, it verifies the consistency
	 * of the repository and it manages a cold recovery if the FOS had not
	 * cleanly stop.
	 */
	public synchronized void start() throws FosException {
		if (loggerFactory == null)
			throw new FosException("MonologFactory required to execute a FosManager.");
		if (dbDir == null)
			throw new FosException("DbDir required to execute a FosManager.");
		if (!new File(dbDir).exists())
			throw new FosException("DbDir [" + dbDir + "] must exist for activating FOS.");
		logger = getLogger(FosLoggerFactory.TXCONTEXT, true);
		entityLogger = getLogger(FosLoggerFactory.TXCONTEXT, false);
		txId = 0;
		txList = new ArrayList();
		File txdir = new File(dbDir + File.separator + TXDIR);
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG, "START operation: dbdir=\"" + dbDir + "\"");
		if (txdir.exists()) {
			if (FosLoggerFactory.DEBUG)
				logger.log(BasicLevel.DEBUG, "TX directory already exists.");
			if (!txdir.isDirectory())
				throw new FosException("TxDir not a directory - see: " + txdir.getPath());
			// Performs the cold recovery phase
			String[] txids = txdir.list();
			FosTxContext txc;
			Xid xid;
			for (int i = 0; i < txids.length; i++) {
				txc = new FosTxContext(entityLogger, this);
				xid = txc.recover(dbDir + File.separator + TXDIR + File.separator
					+ txids[i]);
				if (xid != null) {
					if (FosLoggerFactory.DEBUG)
						logger.log(BasicLevel.DEBUG, "Need DTP recovery for TX: " + xid);
					if (xidList == null)
						xidList = new ArrayList();
					xidList.add(xid);
				}
			}
		} else {
			// First activation: initializes TXDIR
			if (FosLoggerFactory.DEBUG)
				logger.log(BasicLevel.DEBUG, "TX directory does not exist: creates it.");
			txdir.mkdir();
		}
		logger.log(BasicLevel.INFO,
			"FosTxContextFactory started: dbDir is <" + dbDir + ">.");
	}

	/**
	 * Stops a FosTxContextFactory.
	 */
	public synchronized void stop() throws FosException {
		if (txList == null)
			throw new FosException("FOS manager has not been started.");
		if (txList.size() > 0)
			throw new FosException("There are still transactions to be released.");
		txList = null;
	}

	/**
	 * Delivers a new transaction identifier.
	 */
	String getTxId() {
		String res = dbDir + File.separator + TXDIR + File.separator
			+ Long.toString(txId);
		if (txId == Long.MAX_VALUE)
			txId = 0;
		else
			txId++;
		return res;
	}

	/**
	 * Recursively deletes the content of a directory.
	 */
	void deleteDirContent(File dir) throws Exception {
		File lf[] = dir.listFiles();
		for (int i = 0; i < lf.length; i++) {
			if (lf[i].isDirectory()) {
				deleteDirContent(lf[i]);
			}
			lf[i].delete();
		}
	}

	/**
	 * Activates or deactivates a transaction wrt to the number of active
	 * transactions.
	 * @param activate	If true, handles a new active transaction. Otherwise,
	 * 					handles the inactivation of a transaction.
	 */
	synchronized void setActiveTX(boolean activate) {
		if (activate)
			nbActiveTx++;
		else
			nbActiveTx--;
	}

	/**
	 * Gets the number of active transactions.
	 */
	int getNbActiveTX() {
		return nbActiveTx;
	}

	// IMPLEMENTATION OF METHODS FROM THE FosManager INTERFACE

	/**
	 * Initializes the root directory for storing files representing data
	 * objects. It removes all existing data files and creates all data for
	 * managing transactions.
	 */
	public synchronized void initialize() throws FosException {
		if (loggerFactory == null)
			throw new FosException("MonologFactory required to execute a FosManager.");
		if (dbDir == null)
			throw new FosException("DbDir required to execute a FosManager.");
		File dbdf = new File(dbDir);
		if (!dbdf.exists())
			throw new FosException("DbDir [" + dbDir + "] must exist for initializing FOS.");
		logger = getLogger(FosLoggerFactory.TXCONTEXT, true);
		entityLogger = getLogger(FosLoggerFactory.TXCONTEXT, false);
		File txdir = new File(dbDir + File.separator + TXDIR);
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG, "INITIALIZE operation: dbdir=\"" + dbDir + "\"");
		try {
			File lf[] = dbdf.listFiles();
			for (int i = 0; i < lf.length; i++) {
				if (lf[i].isDirectory()) {
					deleteDirContent(lf[i]);
				}
				lf[i].delete();
			}
		} catch (Exception e) {
			throw new FosException("Problem during DbDir initialization.", e);
		}
	}

	/**
	 * Gets all the Xid of distributed transaction that need to be recovered
	 * by a DTP monitor.
	 * @return	The array of Xid of transactions that need to be recovered.
	 */
	public Xid[] getXidForRecovery() {
		Xid[] res;
		if (xidList == null)
			res = new Xid[0];
		res = (Xid[]) xidList.toArray();
		xidList = null;
		return res;
	}

	/**
	 * Creates a new TxContext for managing transactional I/Os with Data Object
	 * Files.
	 */
	public synchronized FosTransaction createTxContext() throws FosException {
		if (xidList != null)
			throw new FosException("Cannot use FOS while distributed TX need to be recovered.");
		FosTransaction res = new FosTxContext(entityLogger, this);
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG, "Creates a new TxContext: " + res);
		txList.add(res);
		return res;
	}

	/**
	 * Releases a FosTxContext. Useful if adding a context pool.
	 * @param txc	The FosTxContext to be released.
	 */
	public synchronized void releaseTxContext(FosTransaction txc) throws FosException {
		if (FosLoggerFactory.DEBUG)
			logger.log(BasicLevel.DEBUG, "Releases TxContext: " + txc);
		if (!txList.contains(txc))
			throw new FosException("Transaction not created with this factory.");
		if (txc.isActive()) {
			if (!((FosTxContext) txc).isPrepared()) {
				logger.log(BasicLevel.WARN, "Releases active transaction: " + txc
					+ " - forces rollback!!");
				txc.rollback();
			} else {
				if (((FosTxContext) txc).getXid() == null) {
					logger.log(BasicLevel.WARN, "Releases prepared transaction: "
						+ txc + " - forces commit!!");
					txc.commit();
				} else
					throw new FosException("Releases a prepared DTP transaction.");
			}
		}
		txList.remove(txc);
	}

	// IMPLEMENTATION OF METHODS FROM THE FosAttributeControler INTERFACE

	/**
	 * Gives access to the FOS root directory (named dbDir).
	 */
	public String getDbDir() {
		return dbDir;
	}

	/**
	 * Changes the FOS root directory.
	 */
	public void setDbDir(String dbdir) {
		dbDir = dbdir;
	}


	// IMPLEMENTATION OF METHODS FROM THE FosLoggerFactory INTERFACE

	/**
	 * Gives access to the logger associated to a given kind of entities managed
	 * by this FOS adapter, which are further classified into factories or
	 * non-factories.
	 * @param entity	The kind of entity (see CONNECTION, MANAGEDCONNECTION,
	 * 					XARESOURCE and TXCONTEXT constants defined above).
	 * @param factory	Specifies if the concerned entity is a factory.
	 * @return	The relevant logger.
	 */
	public Logger getLogger(byte entity, boolean factory) {
		String res = "org.objectweb.perseus.jdo.fos[dbDir=" + dbDir + "]";
		if (factory)
			res += ".fact";
		else
			res += ".inst";
		switch (entity) {
		case FosLoggerFactory.CONNECTION:
			res += ".connx";
			break;
		case FosLoggerFactory.MANAGEDCONNECTION:
			res += ".mancn";
			break;
		case FosLoggerFactory.XARESOURCE:
			res += ".xares";
			break;
		case FosLoggerFactory.TXCONTEXT:
			res += ".txctx";
			break;
		default :
			;
		}
		return loggerFactory.getLogger(res);
	}

	public void setMonologFactory(MonologFactory mf) {
		loggerFactory = mf;
	}
}
