/**
 * ObjectWeb Connector: an implementation of JCA Sun specification along
 *                      with some extensions of this specification.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 0.1
 *
 * Contact: jorm@objectweb.org
 *
 * Author: S. Chassande-Barrioz, P. Dechamboux
 *
 */

package org.objectweb.perseus.connector.ra;

import junit.framework.TestCase;

import java.util.Enumeration;
import java.util.Iterator;
import java.util.Vector;
import javax.naming.Context;
import javax.resource.NotSupportedException;
import javax.resource.ResourceException;
import javax.resource.cci.Connection;
import javax.resource.cci.ConnectionFactory;
import javax.resource.cci.LocalTransaction;
import javax.resource.spi.ManagedConnectionFactory;

/**
 * The class <b>TestRA</b> defines a test suite for testing a correct RA
 */
public abstract class TestRA extends TestCase {
	/**
	 * The managed connection factory of the resource adapter
	 */
	protected ManagedConnectionFactory mcf = null;

	/**
	 * The connection factory of the resource adapter
	 */
	protected ConnectionFactory cf = null;

	/**
	 * The list of writing tests. This vector contains a list of Writing
	 * elements. A writing element permit to abstract the accesses to the
	 * persitent support.
	 */
	protected Vector writings = null;

	/**
	 * Constructor in according to the JUnit constraints
	 */
	protected TestRA(String tn) {
		super(tn);
		writings = getWritings();
	}


	//////////////////////
	// ABSTRACT METHODS //
	//////////////////////

	/**
	 * The method implementation must return a new instance of the resource
	 * adapter. The ManagedConnectionFactory is the identifer of the
	 * resource adapter.
	 */
	protected abstract ManagedConnectionFactory getNewMCFInstance()
		throws Exception;

	/**
	 * The method implementation must initialize the resource adapter
	 */
	protected abstract void initMCF() throws Exception;

	protected abstract void endMCF() throws Exception;

	/**
	 * This method must return a list of writing test.
	 * The return type must be a Vector of Writing instances
	 */
	protected abstract Vector getWritings();


	/////////////////////////////////////////////////////
	// IMPLEMENTATION OF METHODS OF THE TestCase class //
	/////////////////////////////////////////////////////

	/**
	 * Initialisation sequence executed before each test case.
	 */
	protected void setUp() {
		try {
			mcf = getNewMCFInstance();
			initMCF();
			cf = (ConnectionFactory) mcf.createConnectionFactory();
		} catch (Exception e) {
			System.out.println("Error during setUp");
			e.printStackTrace();
//			System.exit(0);
		}
	}

	/**
	 * Termination sequence executed after each test case.
	 */
	protected void tearDown() {
		try {
			endMCF();
		} catch (Exception e) {
			System.out.println("Error during tearDown");
			e.printStackTrace();
//			System.exit(0);
		}
		mcf = null;
		cf = null;
	}


	//////////////////////////////////
	// TESTS OF THE Client CONTRACT //
	//////////////////////////////////

	/**
	 * This simple test check if it is possible to fetch a Connection,
	 * then to close it
	 */
	public void testGetnClose() {
		try {
			Connection c = cf.getConnection();
			c.close();
		} catch (Exception e) {
			e.printStackTrace();
			fail();
		}
	}

	/**
	 * This test, fetches a Connection, starts a local transaction, passes all
	 * writings, and commits the transaction
	 */
	public void testLTCommit() {
		try {
			Connection c = cf.getConnection();
			Enumeration e = writings.elements();
			Writing w;
			while (e.hasMoreElements()) {
				w = (Writing) e.nextElement();
				w.setConnection(c);
				c.setAutoCommit(true);
				w.init();
				if (!w.read(false)) {
					System.err.println("Fail-1!!!");
					fail();
				}
				c.setAutoCommit(false);
				LocalTransaction lt = c.getLocalTransaction();
				lt.begin();
				if (w.read(true)) {
					System.err.println("Fail-2!!!");
					fail();
				}
				w.write();
				if (w.read(false)) {
					System.err.println("Fail-3!!!");
					fail();
				}
				lt.commit();
			}
			c.close();

			c = cf.getConnection();
			e = writings.elements();
			while (e.hasMoreElements()) {
				w = (Writing) e.nextElement();
				w.setConnection(c);
				if (w.read(false)) {
					System.err.println("Fail-4!!!");
					fail();
				}
			}
			c.close();
		} catch (Exception e) {
			e.printStackTrace();
			fail();
		}
	}

	/**
	 * This test, fetches a Connection, starts a local transaction, passes all
	 * writings, and rollbackes the transaction
	 */
	public void testLTRollback() {
		try {
			Connection c = cf.getConnection();
			Enumeration e = writings.elements();
			Writing w = null;
			while (e.hasMoreElements()) {
				w = (Writing) e.nextElement();
				w.setConnection(c);
				c.setAutoCommit(true);
				w.init();
				if (!w.read(false)) {
					System.err.println("Fail-1!!!");
					fail();
				}
				c.setAutoCommit(false);
				LocalTransaction lt = c.getLocalTransaction();
				lt.begin();
				if (w.read(true)) {
					System.err.println("Fail-2!!!");
					fail();
				}
				w.write();
				if (w.read(false)) {
					System.err.println("Fail-3!!!");
					fail();
				}
				lt.rollback();
			}
			c.close();

			c = cf.getConnection();
			e = writings.elements();
			while (e.hasMoreElements()) {
				w = (Writing) e.nextElement();
				w.setConnection(c);
				if (!w.read(false)) {
					System.err.println("Fail-4!!!");
					fail();
				}
			}
			c.close();
		} catch (Exception e) {
			e.printStackTrace();
			fail();
		}
	}

	/**
	 * Check if the meta information of the ConnectionFactory is
	 * reachable and not null.
	 */
	public void testCFMetaData() {
		try {
			try {
				cf.getMetaData().getAdapterName();
			} catch (NotSupportedException nse) {
				return;
			}
			cf.getMetaData().getAdapterName();
			cf.getMetaData().getAdapterShortDescription();
			cf.getMetaData().getAdapterVendorName();
			cf.getMetaData().getAdapterVersion();
			cf.getMetaData().getSpecVersion();
			cf.getMetaData().supportsExecuteWithInputAndOutputRecord();
			cf.getMetaData().supportsExecuteWithInputRecordOnly();
			cf.getMetaData().supportsLocalTransactionDemarcation();
		} catch (Exception e) {
			fail();
		}
	}

	/**
	 * This test checks if the record factory is reachable
	 */
	public void testCFRecordFactory() {
		try {
			cf.getRecordFactory().createIndexedRecord("test1");
		} catch (NotSupportedException nse) {
			// OK
		} catch (Exception e) {
			fail();
		}

		try {
			cf.getRecordFactory().createMappedRecord("test2");
		} catch (NotSupportedException nse) {
			// OK
		} catch (Exception e) {
			fail();
		}
	}

	/**
	 *
	 */
	public void testAutoCommitFalse() {
		System.out.println("\n** testAutoCommitFalse doesn't run !");
		/*
		System.out.println("testAutoCommitFalse");
		try {
			Connection c  = cf.getConnection();
			// Initialise and write data
			Enumeration e = writings.elements();
			Writing w = null;
			while ( e.hasMoreElements() ) {
				w = (Writing) e.nextElement();
				System.out.println("w : "+w);
				w.setConnection(c);
				c.setAutoCommit(true);
				assert( c.getAutoCommit() );
				w.init();
				assert( w.read(false) );
				System.out.println("\tinit,read Ok");
				c.setAutoCommit(false);
				// Check the auto commit state
				assert( ! c.getAutoCommit() );
				w.write();
				assert( w.read(true) );
				System.out.println("\twrite Ok");
			}
		System.out.println("--------Check---------");

			// Check that the writing are not kept
			Connection c2  = cf.getConnection();
			e = writings.elements();
			while ( e.hasMoreElements() ) {
				w = (Writing) e.nextElement();
				System.out.println("w : "+w);
				assert( w.read(c2, false) );
				System.out.println("\tread Ok");
			}
			c.close();
			c2.close();
		} catch (NotSupportedException nse) {
			assert(true);
		} catch (Exception re) {
			assert(false);
		}
		*/
	}

	/**
	 *
	 */
	public void testAutoCommitTrue() {
		try {
			Connection c = cf.getConnection();
			c.setAutoCommit(true);
			// Check the auto commit state
			if (!c.getAutoCommit())
				fail();
			// Initialise and write data
			Enumeration e = writings.elements();
			Writing w = null;
			while (e.hasMoreElements()) {
				w = (Writing) e.nextElement();
				w.init(c);
				if (!w.read(c, false))
					fail();
				w.write(c);
				if (!w.read(c, true))
					fail();
			}
			c.close();

			// Check that the writing are not kept
			c = cf.getConnection();
			e = writings.elements();
			w = null;
			while (e.hasMoreElements()) {
				w = (Writing) e.nextElement();
				if (!w.read(c, true))
					fail();
			}
			c.close();
		} catch (NotSupportedException nse) {
			// OK
		} catch (Exception e) {
			fail();
		}
	}
}