/**
 * ObjectWeb Connector: an implementation of JCA Sun specification along
 *                      with some extensions of this specification.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 0.1
 *
 * Contact: jorm@objectweb.org
 *
 * Author: S. Chassande-Barrioz, P. Dechamboux
 *
 */

package org.objectweb.perseus.connector.ra.fos;

import org.objectweb.perseus.connector.ra.Writing;
import org.objectweb.perseus.fos.api.FosAccess;
import org.objectweb.perseus.fos.api.FosException;
import org.objectweb.perseus.fos.api.FosStructure;

import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import javax.resource.cci.Connection;

/**
 * This class is an implementation of the Writing interface for the FOS case.
 * This class accesses to a directory that stores objects whose identifier is
 * pid and value is description.
 */
public class FosWriting implements Writing, FosStructure {
	private static final String OBJDIR = "objdir";
	private String pid;
	private String details;
	private Connection connection;
	private FosAccess lastUsedInFS;

	protected FosWriting(int _pid, String _details) {
		pid = Integer.toString(_pid);
		details = _details;
	}

	public void init(Connection co) throws Exception {
		FosAccess fa = (FosAccess) co;
		if (fa.exist(OBJDIR, pid))
			fa.delete(OBJDIR, pid);
		fa.write(OBJDIR, pid, this, null);
	}

	public void init() throws Exception {
		init(connection);
	}

	/**
	 * Tests if connection is correctly passed through to FosStructure.
	 * @return	false if it is not passed through correctly.
	 */
	public boolean passThrough(Connection co) {
		try {
			FosAccess fa = (FosAccess) co;
			lastUsedInFS = null;
			fa.write(OBJDIR, pid, this, null);
			if (lastUsedInFS != fa)
				return false;
			lastUsedInFS = null;
			fa.read(OBJDIR, pid, this, null);
			if (lastUsedInFS != fa)
				return false;
			return true;
		} catch (FosException fe) {
			fe.printStackTrace();
			return false;
		}
	}

	/**
	 * Read a data from the support, and check the value. The boolean
	 * parameter is used to check the read value.
	 * The method returns true is the data state is right
	 * The method throws Exceptions if an error occures during the test
	 */
	public boolean read(Connection co, boolean iswritten)
		throws Exception {
		FosAccess fa = (FosAccess) co;
		return fa.exist(OBJDIR, pid) ^ iswritten;
	}

	public boolean read(boolean iswritten)
		throws Exception {
		return read(connection, iswritten);
	}

	/**
	 * The method throws Exceptions if an error occures during the test
	 */
	public void write(Connection co) throws Exception {
		FosAccess fa = (FosAccess) co;
		if (fa.exist(OBJDIR, pid))
			fa.delete(OBJDIR, pid);
	}

	/**
	 * The method throws Exceptions if an error occures during the test
	 */
	public void write() throws Exception {
		write(connection);
	}

	public void setConnection(Connection c) throws Exception {
		connection = c;
	}

	/**
	 * Reads the content of a persistent object from the file represented by
	 * the given ObjectInputStream.
	 * @param ois	The ObjectInputStream representing the file from which to
	 *				read.
	 * @param conn	The connection used to access the File Object Store.
	 */
	public void readFile(ObjectInputStream ois, FosAccess conn, Object ctxt) throws Exception {
		details = ois.readUTF();
		lastUsedInFS = conn;
	}

	/**
	 * Writes the content of a persistent object to the file represented by
	 * the given ObjectOutputStream.
	 * @param oos	The ObjectOutputStream representing the file to which to
	 *				write.
	 * @param conn	The connection used to access the File Object Store.
	 */
	public void writeFile(ObjectOutputStream oos, FosAccess conn, Object ctxt) throws Exception {
		oos.writeUTF(details);
		lastUsedInFS = conn;
	}
}