/**
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 */

package org.objectweb.perseus.fos;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;
import junit.textui.TestRunner;
import org.objectweb.perseus.LogBoot;
import org.objectweb.perseus.fos.api.FosAccess;
import org.objectweb.perseus.fos.api.FosException;
import org.objectweb.perseus.fos.api.FosManager;
import org.objectweb.perseus.fos.api.FosStructure;
import org.objectweb.perseus.fos.api.FosTransaction;
import org.objectweb.perseus.fos.lib.FosTxContextFactory;

import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.Properties;

/**
 * Tests standard transactions that correctly executes (correct begin/commit
 * sequences).
 */
public class TestRecovery extends TestCase {
	private static final int INITIALVALUE = 10;
	private static final int NEWVALUE = 20;
	private static final int NULLVALUE = 0;
	private FosManager fm;
	private static final String POBJDIR = "c1";
	private String[] pobjId = {
		"obj1", "obj2", "obj3", "obj4", "obj5"
	};
	private String dbDir;
	private IntPObj[] pobjVal = {
		new IntPObj(),
		new IntPObj(),
		new IntPObj(),
		new IntPObj(),
		new IntPObj()
	};

	// IMPLEMENTATION OF METHODS OF THE TestCase CLASS

	/**
	 * Initialisation sequence executed before each test case.
	 */
	protected void setUp() {
		try {
			fm = new FosTxContextFactory();
			fm.setDbDir(dbDir);
			fm.setMonologFactory(LogBoot.getLoggerFactory());
			fm.initialize();
		} catch (FosException fe) {
			System.out.println("Error during setUp - 1");
			if (fe.getNestedException() != null)
				fe.getNestedException().printStackTrace();
			fe.printStackTrace();
			System.exit(0);
		} catch (Exception e) {
			System.out.println("Error during setUp - 2");
			e.printStackTrace();
			System.exit(0);
		}
	}

	/**
	 * Termination sequence executed after each test case.
	 */
	protected void tearDown() {
	}

	// Tests transaction operations (no data manipulation).

	/**
	 * Tests start of an empty FOS repository.
	 */
	public void testRecoverInitialized() {
		try {
			fm = new FosTxContextFactory();
			fm.setDbDir(dbDir);

			fm.setMonologFactory(LogBoot.getLoggerFactory());
			fm.start();
			fm = new FosTxContextFactory();
			fm.setDbDir(dbDir);

			fm.setMonologFactory(LogBoot.getLoggerFactory());
			fm.start();
			fm.stop();
		} catch (FosException fe) {
			fe.printStackTrace();
			fail("FOS error.");
		} catch (Exception e) {
			e.printStackTrace();
			fail("Log configuration error.");
		}
	}

	/**
	 * Tests start of a FOS repository with a committed creation.
	 */
	public void testRecoverCommitCreate() {
		try {
			fm = new FosTxContextFactory();
			fm.setDbDir(dbDir);

			fm.setMonologFactory(LogBoot.getLoggerFactory());
			fm.start();
			FosTransaction ft = fm.createTxContext();
			ft.begin();
			ft.write(POBJDIR, pobjId[0], pobjVal[0], null);
			ft.commit();
			fm = new FosTxContextFactory();
			fm.setDbDir(dbDir);

			fm.setMonologFactory(LogBoot.getLoggerFactory());
			fm.start();
			ft.begin();
			boolean ex = ft.exist(POBJDIR, pobjId[0]);
			ft.commit();
			fm.stop();
			if (!ex)
				fail();
		} catch (FosException fe) {
			fe.printStackTrace();
			fail("FOS error.");
		} catch (Exception e) {
			e.printStackTrace();
			fail("Log configuration error.");
		}
	}

	/**
	 * Tests start of a FOS repository with a prepared creation.
	 */
	public void testRecoverPrepareCreate() {
		try {
			fm = new FosTxContextFactory();
			fm.setDbDir(dbDir);

			fm.setMonologFactory(LogBoot.getLoggerFactory());
			fm.start();
			FosTransaction ft = fm.createTxContext();
			ft.begin();
			ft.write(POBJDIR, pobjId[0], pobjVal[0], null);
			ft.prepare();
			fm = new FosTxContextFactory();
			fm.setDbDir(dbDir);

			fm.setMonologFactory(LogBoot.getLoggerFactory());
			fm.start();
			ft.begin();
			boolean ex = ft.exist(POBJDIR, pobjId[0]);
			ft.commit();
			fm.stop();
			if (!ex)
				fail();
		} catch (FosException fe) {
			fe.printStackTrace();
			fail("FOS error.");
		} catch (Exception e) {
			e.printStackTrace();
			fail("Log configuration error.");
		}
	}

	/**
	 * Tests start of a FOS repository with a rolled back creation.
	 */
	public void testRecoverRollbackCreate() {
		try {
			fm = new FosTxContextFactory();
			fm.setDbDir(dbDir);

			fm.setMonologFactory(LogBoot.getLoggerFactory());
			fm.start();
			FosTransaction ft = fm.createTxContext();
			ft.begin();
			ft.write(POBJDIR, pobjId[0], pobjVal[0], null);
			ft.rollback();
			fm = new FosTxContextFactory();
			fm.setDbDir(dbDir);

			fm.setMonologFactory(LogBoot.getLoggerFactory());
			fm.start();
			ft.begin();
			boolean ex = ft.exist(POBJDIR, pobjId[0]);
			ft.commit();
			fm.stop();
			if (ex)
				fail();
		} catch (FosException fe) {
			fe.printStackTrace();
			fail("FOS error.");
		} catch (Exception e) {
			e.printStackTrace();
			fail("Log configuration error.");
		}
	}

	/**
	 * Tests start of a FOS repository with a uncommitted creation.
	 */
	public void testRecoverCreate() {
		try {
			fm = new FosTxContextFactory();
			fm.setDbDir(dbDir);

			fm.setMonologFactory(LogBoot.getLoggerFactory());
			fm.start();
			FosTransaction ft = fm.createTxContext();
			ft.begin();
			ft.write(POBJDIR, pobjId[0], pobjVal[0], null);
			fm = new FosTxContextFactory();
			fm.setDbDir(dbDir);

			fm.setMonologFactory(LogBoot.getLoggerFactory());
			fm.start();
			ft.begin();
			boolean ex = ft.exist(POBJDIR, pobjId[0]);
			ft.commit();
			fm.stop();
			if (ex)
				fail();
		} catch (FosException fe) {
			fe.printStackTrace();
			fail("FOS error.");
		} catch (Exception e) {
			e.printStackTrace();
			fail("Log configuration error.");
		}
	}

	/**
	 * Tests start of a FOS repository with a committed deletion.
	 */
	public void testRecoverCommitDelete() {
		try {
			fm = new FosTxContextFactory();
			fm.setDbDir(dbDir);

			fm.setMonologFactory(LogBoot.getLoggerFactory());
			fm.start();
			FosTransaction ft = fm.createTxContext();
			ft.begin();
			ft.write(POBJDIR, pobjId[0], pobjVal[0], null);
			ft.commit();
			ft.begin();
			ft.delete(POBJDIR, pobjId[0]);
			ft.commit();
			fm = new FosTxContextFactory();
			fm.setDbDir(dbDir);

			fm.setMonologFactory(LogBoot.getLoggerFactory());
			fm.start();
			ft.begin();
			boolean ex = ft.exist(POBJDIR, pobjId[0]);
			ft.commit();
			fm.stop();
			if (ex)
				fail();
		} catch (FosException fe) {
			fe.printStackTrace();
			fail("FOS error.");
		} catch (Exception e) {
			e.printStackTrace();
			fail("Log configuration error.");
		}
	}

	/**
	 * Tests start of a FOS repository with a prepared deletion.
	 */
	public void testRecoverPrepareDelete() {
		try {
			fm = new FosTxContextFactory();
			fm.setDbDir(dbDir);

			fm.setMonologFactory(LogBoot.getLoggerFactory());
			fm.start();
			FosTransaction ft = fm.createTxContext();
			ft.begin();
			ft.write(POBJDIR, pobjId[0], pobjVal[0], null);
			ft.commit();
			ft.begin();
			ft.delete(POBJDIR, pobjId[0]);
			ft.prepare();
			fm = new FosTxContextFactory();
			fm.setDbDir(dbDir);

			fm.setMonologFactory(LogBoot.getLoggerFactory());
			fm.start();
			ft.begin();
			boolean ex = ft.exist(POBJDIR, pobjId[0]);
			ft.commit();
			fm.stop();
			if (ex)
				fail();
		} catch (FosException fe) {
			fe.printStackTrace();
			fail("FOS error.");
		} catch (Exception e) {
			e.printStackTrace();
			fail("Log configuration error.");
		}
	}

	/**
	 * Tests start of a FOS repository with a rolled back deletion.
	 */
	public void testRecoverRollbackDelete() {
		try {
			fm = new FosTxContextFactory();
			fm.setDbDir(dbDir);

			fm.setMonologFactory(LogBoot.getLoggerFactory());
			fm.start();
			FosTransaction ft = fm.createTxContext();
			ft.begin();
			ft.write(POBJDIR, pobjId[0], pobjVal[0], null);
			ft.commit();
			ft.begin();
			ft.delete(POBJDIR, pobjId[0]);
			ft.rollback();
			fm = new FosTxContextFactory();
			fm.setDbDir(dbDir);

			fm.setMonologFactory(LogBoot.getLoggerFactory());
			fm.start();
			ft.begin();
			boolean ex = ft.exist(POBJDIR, pobjId[0]);
			ft.commit();
			fm.stop();
			if (!ex)
				fail();
		} catch (FosException fe) {
			fe.printStackTrace();
			fail("FOS error.");
		} catch (Exception e) {
			e.printStackTrace();
			fail("Log configuration error.");
		}
	}

	/**
	 * Tests start of a FOS repository with a uncommitted deletion.
	 */
	public void testRecoverDelete() {
		try {
			fm = new FosTxContextFactory();
			fm.setDbDir(dbDir);

			fm.setMonologFactory(LogBoot.getLoggerFactory());
			fm.start();
			FosTransaction ft = fm.createTxContext();
			ft.begin();
			ft.write(POBJDIR, pobjId[0], pobjVal[0], null);
			ft.commit();
			ft.begin();
			ft.delete(POBJDIR, pobjId[0]);
			fm = new FosTxContextFactory();
			fm.setDbDir(dbDir);

			fm.setMonologFactory(LogBoot.getLoggerFactory());
			fm.start();
			ft.begin();
			boolean ex = ft.exist(POBJDIR, pobjId[0]);
			ft.commit();
			fm.stop();
			if (!ex)
				fail();
		} catch (FosException fe) {
			fe.printStackTrace();
			fail("FOS error.");
		} catch (Exception e) {
			e.printStackTrace();
			fail("Log configuration error.");
		}
	}

	/**
	 * Tests start of a FOS repository with a committed modification.
	 */
	public void testRecoverCommitModify() {
		try {
			fm = new FosTxContextFactory();
			fm.setDbDir(dbDir);

			fm.setMonologFactory(LogBoot.getLoggerFactory());
			fm.start();
			FosTransaction ft = fm.createTxContext();
			ft.begin();
			ft.write(POBJDIR, pobjId[0], pobjVal[0], null);
			ft.commit();
			ft.begin();
			pobjVal[0].setValue(NEWVALUE);
			ft.write(POBJDIR, pobjId[0], pobjVal[0], null);
			ft.commit();
			fm = new FosTxContextFactory();
			fm.setDbDir(dbDir);

			fm.setMonologFactory(LogBoot.getLoggerFactory());
			fm.start();
			pobjVal[0].setValue(NULLVALUE);
			ft.begin();
			ft.read(POBJDIR, pobjId[0], pobjVal[0], null);
			ft.commit();
			fm.stop();
			if (pobjVal[0].getValue() != NEWVALUE)
				fail();
		} catch (FosException fe) {
			fe.printStackTrace();
			fail("FOS error.");
		} catch (Exception e) {
			e.printStackTrace();
			fail("Log configuration error.");
		}
	}

	/**
	 * Tests start of a FOS repository with a prepared modification.
	 */
	public void testRecoverPrepareModify() {
		try {
			fm = new FosTxContextFactory();
			fm.setDbDir(dbDir);

			fm.setMonologFactory(LogBoot.getLoggerFactory());
			fm.start();
			FosTransaction ft = fm.createTxContext();
			ft.begin();
			ft.write(POBJDIR, pobjId[0], pobjVal[0], null);
			ft.commit();
			ft.begin();
			pobjVal[0].setValue(NEWVALUE);
			ft.write(POBJDIR, pobjId[0], pobjVal[0], null);
			ft.prepare();
			fm = new FosTxContextFactory();
			fm.setDbDir(dbDir);

			fm.setMonologFactory(LogBoot.getLoggerFactory());
			fm.start();
			pobjVal[0].setValue(NULLVALUE);
			ft.begin();
			ft.read(POBJDIR, pobjId[0], pobjVal[0], null);
			ft.commit();
			fm.stop();
			if (pobjVal[0].getValue() != NEWVALUE)
				fail();
		} catch (FosException fe) {
			fe.printStackTrace();
			fail("FOS error.");
		} catch (Exception e) {
			e.printStackTrace();
			fail("Log configuration error.");
		}
	}

	/**
	 * Tests start of a FOS repository with a rolled back modification.
	 */
	public void testRecoverRollbackModify() {
		try {
			fm = new FosTxContextFactory();
			fm.setDbDir(dbDir);

			fm.setMonologFactory(LogBoot.getLoggerFactory());
			fm.start();
			FosTransaction ft = fm.createTxContext();
			ft.begin();
			ft.write(POBJDIR, pobjId[0], pobjVal[0], null);
			ft.commit();
			ft.begin();
			pobjVal[0].setValue(NEWVALUE);
			ft.write(POBJDIR, pobjId[0], pobjVal[0], null);
			ft.rollback();
			fm = new FosTxContextFactory();
			fm.setDbDir(dbDir);

			fm.setMonologFactory(LogBoot.getLoggerFactory());
			fm.start();
			pobjVal[0].setValue(NULLVALUE);
			ft.begin();
			ft.read(POBJDIR, pobjId[0], pobjVal[0], null);
			ft.commit();
			fm.stop();
			if (pobjVal[0].getValue() != INITIALVALUE)
				fail();
		} catch (FosException fe) {
			fe.printStackTrace();
			fail("FOS error.");
		} catch (Exception e) {
			e.printStackTrace();
			fail("Log configuration error.");
		}
	}

	/**
	 * Tests start of a FOS repository with a uncommitted modification.
	 */
	public void testRecoverModify() {
		try {
			fm = new FosTxContextFactory();
			fm.setDbDir(dbDir);

			fm.setMonologFactory(LogBoot.getLoggerFactory());
			fm.start();
			FosTransaction ft = fm.createTxContext();
			ft.begin();
			ft.write(POBJDIR, pobjId[0], pobjVal[0], null);
			ft.commit();
			ft.begin();
			pobjVal[0].setValue(NEWVALUE);
			ft.write(POBJDIR, pobjId[0], pobjVal[0], null);
			fm = new FosTxContextFactory();
			fm.setDbDir(dbDir);

			fm.setMonologFactory(LogBoot.getLoggerFactory());
			fm.start();
			pobjVal[0].setValue(NULLVALUE);
			ft.begin();
			ft.read(POBJDIR, pobjId[0], pobjVal[0], null);
			ft.commit();
			fm.stop();
			if (pobjVal[0].getValue() != INITIALVALUE)
				fail();
		} catch (FosException fe) {
			fe.printStackTrace();
			fail("FOS error.");
		} catch (Exception e) {
			e.printStackTrace();
			fail("Log configuration error.");
		}
	}

	// For managing the TestCase as a standalone Test Suite.

	/**
	 * Constructs the suite containing all test cases defined within
	 * TestRecovery.
	 */
	public static Test suite() {
		return new TestSuite(TestRecovery.class);
	}

	/**
	 * For running the TestRecovery suite standalone.
	 */
	public static void main(String args[]) {
		TestRunner.run(suite());
	}

	/**
	 * Constructs a TestRecovery.
	 */
	public TestRecovery(String tn) throws Exception {
		super(tn);
		Properties prop = new Properties();
		prop.load(ClassLoader.getSystemResourceAsStream("owfos.properties"));
		dbDir = prop.getProperty("dbdir", null);
		if (dbDir == null) {
			throw new Exception(
				"Cannot find the 'dbdir' property");
		}
	}

	/**
	 *
	 */
	class IntPObj implements FosStructure {
		/**
		 * The value of the integer persistent object.
		 */
		private int value = INITIALVALUE;

		/**
		 * Reads the content of a persistent object from the file represented by
		 * the given ObjectInputStream.
		 * @param ois	The ObjectInputStream representing the file from which to
		 *				read.
		 * @param conn	The connection used to access the File Object Store.
		 */
		public void readFile(ObjectInputStream ois, FosAccess conn, Object ctxt) throws Exception {
			value = ois.readInt();
		}

		/**
		 * Writes the content of a persistent object to the file represented by
		 * the given ObjectOutputStream.
		 * @param oos	The ObjectOutputStream representing the file to which to
		 *				write.
		 * @param conn	The connection used to access the File Object Store.
		 */
		public void writeFile(ObjectOutputStream oos, FosAccess conn, Object ctxt) throws Exception {
			oos.writeInt(value);
		}

		/**
		 * Returns the value of this integer persistent object.
		 */
		int getValue() {
			return value;
		}

		/**
		 * Assigns the value of this integer persistent object.
		 */
		void setValue(int val) {
			value = val;
		}
	}
}