/**
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 */

package org.objectweb.perseus.fos;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;
import junit.textui.TestRunner;
import org.objectweb.perseus.LogBoot;
import org.objectweb.perseus.fos.api.FosException;
import org.objectweb.perseus.fos.api.FosManager;
import org.objectweb.perseus.fos.api.FosStructure;
import org.objectweb.perseus.fos.api.FosTransaction;
import org.objectweb.perseus.fos.api.FosAccess;
import org.objectweb.perseus.fos.lib.FosTxContextFactory;

import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.Iterator;
import java.util.Properties;

/**
 * Tests standard transactions that correctly executes (correct begin/commit
 * sequences).
 */
public class TestTxOps extends TestCase {
	private static final int INITIALVALUE = 10;
	private static final int NEWVALUE = 20;
	private static final int NULLVALUE = 0;
	FosAccess tx;
	FosManager fm;
	static final String POBJDIR = "c1";
	String[] pobjId = {
		"obj1", "obj2", "obj3", "obj4", "obj5"
	};
	String dbDir;
	IntPObj[] pobjVal = {
		new IntPObj(),
		new IntPObj(),
		new IntPObj(),
		new IntPObj(),
		new IntPObj()
	};

	// IMPLEMENTATION OF METHODS OF THE TestCase CLASS

	/**
	 * Initialisation sequence executed before each test case.
	 */
	protected void setUp() {
		try {
			fm = new FosTxContextFactory();
			fm.setDbDir(dbDir);
			fm.setMonologFactory(LogBoot.getLoggerFactory());
			fm.initialize();
			fm.start();
		} catch (FosException fe) {
			System.out.println("Error during setUp - 1");
			if (fe.getNestedException() != null)
				fe.getNestedException().printStackTrace();
			fe.printStackTrace();
			System.exit(0);
		} catch (Exception e) {
			System.out.println("Error during setUp - 2");
			e.printStackTrace();
			System.exit(0);
		}
	}

	/**
	 * Termination sequence executed after each test case.
	 */
	protected void tearDown() {
		try {
			fm.stop();
		} catch (FosException fe) {
			System.out.println("Error during tearDown");
			if (fe.getNestedException() != null)
				fe.getNestedException().printStackTrace();
			fe.printStackTrace();
		}
	}

	// Tests transaction operations (no data manipulation).

	/**
	 * Tests a simple creation and releasing of a transaction that
	 * is correct.
	 */
	public void testCreateRelease() {
		try {
			FosTransaction ft = fm.createTxContext();
			fm.releaseTxContext(ft);
		} catch (FosException fe) {
			fail();
		}
	}

	/**
	 * Tests a simple creation and releasing of a transaction along
	 * with a begin/commit sequence.
	 */
	public void testCRBeginCommit() {
		try {
			FosTransaction ft = fm.createTxContext();
			ft.begin();
			ft.commit();
			fm.releaseTxContext(ft);
		} catch (FosException fe) {
			fail();
		}
	}

	/**
	 * Tests a simple creation and releasing of a transaction along
	 * with a begin/prepare/commit sequence.
	 */
	public void testCRBeginPrepareCommit() {
		try {
			FosTransaction ft = fm.createTxContext();
			ft.begin();
			ft.prepare();
			ft.commit();
			fm.releaseTxContext(ft);
		} catch (FosException fe) {
			fail();
		}
	}

	/**
	 * Tests a simple creation and releasing of a transaction along
	 * with a begin/prepare/rollback sequence.
	 */
	public void testCRBeginPrepareRollback() {
		try {
			FosTransaction ft = fm.createTxContext();
			ft.begin();
			ft.prepare();
			ft.rollback();
			fm.releaseTxContext(ft);
		} catch (FosException fe) {
			fail();
		}
	}

	/**
	 * Tests a simple creation and releasing of a transaction along
	 * with a begin/rollback sequence.
	 */
	public void testCRBeginRollback() {
		try {
			FosTransaction ft = fm.createTxContext();
			ft.begin();
			ft.rollback();
			fm.releaseTxContext(ft);
		} catch (FosException fe) {
			fail();
		}
	}

	/**
	 * Tests a simple creation and releasing of a transaction along
	 * with a simple begin: rollback forced.
	 */
	public void testCRBeginAlone() {
		try {
			FosTransaction ft = fm.createTxContext();
			ft.begin();
			fm.releaseTxContext(ft);
		} catch (FosException fe) {
			fail();
		}
	}

	/**
	 * Tests a simple creation and releasing of a transaction along
	 * with a simple begin/prepare sequence: commit forced.
	 */
	public void testCRBeginPrepare() {
		try {
			FosTransaction ft = fm.createTxContext();
			ft.begin();
			ft.prepare();
			fm.releaseTxContext(ft);
		} catch (FosException fe) {
			fail();
		}
	}

	// Tests FosAccess operations (data manipulations).

	/**
	 * Tests a simple exist test where the object does not exist.
	 */
	public void testExistNoObj() {
		try {
			FosTransaction ft = fm.createTxContext();
			ft.begin();
			boolean ex = ft.exist(POBJDIR, pobjId[0]);
			ft.commit();
			fm.releaseTxContext(ft);
			if (ex)
				fail();
		} catch (FosException fe) {
			if (fe.getNestedException() != null)
				fe.getNestedException().printStackTrace();
			fe.printStackTrace();
			fail();
		}
	}

	/**
	 * Tests commit of a simple creation and exist on an existing object.
	 */
	public void testCommitCreation() {
		try {
			FosTransaction ft = fm.createTxContext();
			ft.begin();
			ft.write(POBJDIR, pobjId[0], pobjVal[0], null);
			ft.commit();
			ft.begin();
			boolean ex = ft.exist(POBJDIR, pobjId[0]);
			ft.commit();
			fm.releaseTxContext(ft);
			if (!ex)
				fail("Object should exist: creation committed!");
		} catch (FosException fe) {
			if (fe.getNestedException() != null)
				fe.getNestedException().printStackTrace();
			fe.printStackTrace();
			fail();
		}
	}

	/**
	 * Tests rollback of a simple creation.
	 */
	public void testRollbackCreation() {
		try {
			FosTransaction ft = fm.createTxContext();
			ft.begin();
			ft.write(POBJDIR, pobjId[0], pobjVal[0], null);
			ft.rollback();
			ft.begin();
			boolean ex = ft.exist(POBJDIR, pobjId[0]);
			ft.commit();
			fm.releaseTxContext(ft);
			if (ex)
				fail("Object should not exist: creation rolled back!");
		} catch (FosException fe) {
			if (fe.getNestedException() != null)
				fe.getNestedException().printStackTrace();
			fe.printStackTrace();
			fail();
		}
	}

	/**
	 * Tests rollback of a simple read.
	 */
	public void testRead() {
		try {
			FosTransaction ft = fm.createTxContext();
			ft.begin();
			ft.write(POBJDIR, pobjId[0], pobjVal[0], null);
			ft.commit();
			pobjVal[0].setValue(NULLVALUE);
			ft.begin();
			ft.read(POBJDIR, pobjId[0], pobjVal[0], null);
			ft.commit();
			fm.releaseTxContext(ft);
			if (pobjVal[0].getValue() != INITIALVALUE)
				fail("Object read: wrong value!");
		} catch (FosException fe) {
			if (fe.getNestedException() != null)
				fe.getNestedException().printStackTrace();
			fe.printStackTrace();
			fail();
		}
	}

	/**
	 * Tests commit of a simple deletion.
	 */
	public void testCommitDeletion() {
		try {
			FosTransaction ft = fm.createTxContext();
			ft.begin();
			ft.write(POBJDIR, pobjId[0], pobjVal[0], null);
			ft.commit();
			ft.begin();
			ft.delete(POBJDIR, pobjId[0]);
			ft.commit();
			ft.begin();
			boolean ex = ft.exist(POBJDIR, pobjId[0]);
			ft.commit();
			fm.releaseTxContext(ft);
			if (ex)
				fail("Object should not exist: deletion committed!");
		} catch (FosException fe) {
			if (fe.getNestedException() != null)
				fe.getNestedException().printStackTrace();
			fe.printStackTrace();
			fail();
		}
	}

	/**
	 * Tests rollback of a simple deletion.
	 */
	public void testRollbackDeletion() {
		try {
			FosTransaction ft = fm.createTxContext();
			ft.begin();
			ft.write(POBJDIR, pobjId[0], pobjVal[0], null);
			ft.commit();
			ft.begin();
			ft.delete(POBJDIR, pobjId[0]);
			ft.rollback();
			ft.begin();
			boolean ex = ft.exist(POBJDIR, pobjId[0]);
			ft.commit();
			fm.releaseTxContext(ft);
			if (!ex)
				fail("Object should exist: deletion rolled back!");
		} catch (FosException fe) {
			if (fe.getNestedException() != null)
				fe.getNestedException().printStackTrace();
			fe.printStackTrace();
			fail();
		}
	}

	/**
	 * Tests commit of a simple modification.
	 */
	public void testCommitModification() {
		try {
			FosTransaction ft = fm.createTxContext();
			ft.begin();
			ft.write(POBJDIR, pobjId[0], pobjVal[0], null);
			ft.commit();
			ft.begin();
			pobjVal[0].setValue(NEWVALUE);
			ft.write(POBJDIR, pobjId[0], pobjVal[0], null);
			ft.commit();
			ft.begin();
			ft.read(POBJDIR, pobjId[0], pobjVal[0], null);
			ft.commit();
			fm.releaseTxContext(ft);
			if (pobjVal[0].getValue() != NEWVALUE)
				fail("Modification committed: wrong value!");
		} catch (FosException fe) {
			if (fe.getNestedException() != null)
				fe.getNestedException().printStackTrace();
			fe.printStackTrace();
			fail();
		}
	}

	/**
	 * Tests rollback of a simple modification.
	 */
	public void testRollbackModification() {
		try {
			FosTransaction ft = fm.createTxContext();
			ft.begin();
			ft.write(POBJDIR, pobjId[0], pobjVal[0], null);
			ft.commit();
			ft.begin();
			pobjVal[0].setValue(NEWVALUE);
			ft.write(POBJDIR, pobjId[0], pobjVal[0], null);
			ft.rollback();
			ft.begin();
			ft.read(POBJDIR, pobjId[0], pobjVal[0], null);
			ft.commit();
			fm.releaseTxContext(ft);
			if (pobjVal[0].getValue() != INITIALVALUE)
				fail("Modification rolled back: wrong value!");
		} catch (FosException fe) {
			if (fe.getNestedException() != null)
				fe.getNestedException().printStackTrace();
			fe.printStackTrace();
			fail();
		}
	}

	/**
	 * Tests commit multiple creations.
	 */
	public void testCommitCreations() {
		try {
			FosTransaction ft = fm.createTxContext();
			ft.begin();
			for (int i = 0; i < pobjId.length; i++)
				ft.write(POBJDIR, pobjId[i], pobjVal[i], null);
			ft.commit();
			ft.begin();
			boolean ex = true;
			for (int i = 0; i < pobjId.length; i++)
				ex &= ft.exist(POBJDIR, pobjId[i]);
			ft.commit();
			fm.releaseTxContext(ft);
			if (!ex)
				fail("Objects should exist: creations committed!");
		} catch (FosException fe) {
			if (fe.getNestedException() != null)
				fe.getNestedException().printStackTrace();
			fe.printStackTrace();
			fail();
		}
	}

	/**
	 * Tests rollback of multiple creations.
	 */
	public void testRollbackCreations() {
		try {
			FosTransaction ft = fm.createTxContext();
			ft.begin();
			for (int i = 0; i < pobjId.length; i++)
				ft.write(POBJDIR, pobjId[i], pobjVal[i], null);
			ft.rollback();
			ft.begin();
			boolean ex = false;
			for (int i = 0; i < pobjId.length; i++)
				ex |= ft.exist(POBJDIR, pobjId[i]);
			ft.commit();
			fm.releaseTxContext(ft);
			if (ex)
				fail("Objects should not exist: creations rolled back!");
		} catch (FosException fe) {
			if (fe.getNestedException() != null)
				fe.getNestedException().printStackTrace();
			fe.printStackTrace();
			fail();
		}
	}

	/**
	 * Tests multiple reads.
	 */
	public void testReads() {
		try {
			FosTransaction ft = fm.createTxContext();
			ft.begin();
			for (int i = 0; i < pobjId.length; i++)
				ft.write(POBJDIR, pobjId[i], pobjVal[i], null);
			ft.commit();
			for (int i = 0; i < pobjId.length; i++)
				pobjVal[i].setValue(NULLVALUE);
			ft.begin();
			for (int i = 0; i < pobjId.length; i++)
				ft.read(POBJDIR, pobjId[i], pobjVal[i], null);
			ft.commit();
			fm.releaseTxContext(ft);
			for (int i = 0; i < pobjId.length; i++)
				if (pobjVal[i].getValue() != INITIALVALUE) {
					fail("Objects read: wrong value!");
					break;
				}
		} catch (FosException fe) {
			if (fe.getNestedException() != null)
				fe.getNestedException().printStackTrace();
			fe.printStackTrace();
			fail();
		}
	}

	/**
	 * Tests commit of multiple deletions.
	 */
	public void testCommitDeletions() {
		try {
			FosTransaction ft = fm.createTxContext();
			ft.begin();
			for (int i = 0; i < pobjId.length; i++)
				ft.write(POBJDIR, pobjId[i], pobjVal[i], null);
			ft.commit();
			ft.begin();
			for (int i = 0; i < pobjId.length; i++)
				ft.delete(POBJDIR, pobjId[i]);
			ft.commit();
			ft.begin();
			boolean nex = true;
			for (int i = 0; i < pobjId.length; i++)
				nex &= !ft.exist(POBJDIR, pobjId[i]);
			ft.commit();
			fm.releaseTxContext(ft);
			if (!nex)
				fail("Objects should not exist: deletions committed!");
		} catch (FosException fe) {
			if (fe.getNestedException() != null)
				fe.getNestedException().printStackTrace();
			fe.printStackTrace();
			fail();
		}
	}

	/**
	 * Tests commit of multiple deletions.
	 */
	public void testRollbackDeletions() {
		try {
			FosTransaction ft = fm.createTxContext();
			ft.begin();
			for (int i = 0; i < pobjId.length; i++)
				ft.write(POBJDIR, pobjId[i], pobjVal[i], null);
			ft.commit();
			ft.begin();
			for (int i = 0; i < pobjId.length; i++)
				ft.delete(POBJDIR, pobjId[i]);
			ft.rollback();
			ft.begin();
			boolean ex = true;
			for (int i = 0; i < pobjId.length; i++)
				ex &= ft.exist(POBJDIR, pobjId[i]);
			ft.commit();
			fm.releaseTxContext(ft);
			if (!ex)
				fail("Objects should exist: deletions rolled back!");
		} catch (FosException fe) {
			if (fe.getNestedException() != null)
				fe.getNestedException().printStackTrace();
			fe.printStackTrace();
			fail();
		}
	}

	/**
	 * Tests commit of multiple modifications.
	 */
	public void testCommitModifications() {
		try {
			FosTransaction ft = fm.createTxContext();
			ft.begin();
			for (int i = 0; i < pobjId.length; i++)
				ft.write(POBJDIR, pobjId[i], pobjVal[i], null);
			ft.commit();
			ft.begin();
			for (int i = 0; i < pobjId.length; i++) {
				pobjVal[i].setValue(NEWVALUE);
				ft.write(POBJDIR, pobjId[i], pobjVal[i], null);
			}
			ft.commit();
			for (int i = 0; i < pobjId.length; i++)
				pobjVal[i].setValue(NULLVALUE);
			ft.begin();
			for (int i = 0; i < pobjId.length; i++)
				ft.read(POBJDIR, pobjId[i], pobjVal[i], null);
			ft.commit();
			fm.releaseTxContext(ft);
			for (int i = 0; i < pobjId.length; i++)
				if (pobjVal[i].getValue() != NEWVALUE) {
					fail("Modifications committed: wrong value!");
					break;
				}
		} catch (FosException fe) {
			if (fe.getNestedException() != null)
				fe.getNestedException().printStackTrace();
			fe.printStackTrace();
			fail();
		}
	}

	/**
	 * Tests rollback of multiple modifications.
	 */
	public void testRollbackModifications() {
		try {
			FosTransaction ft = fm.createTxContext();
			ft.begin();
			for (int i = 0; i < pobjId.length; i++)
				ft.write(POBJDIR, pobjId[i], pobjVal[i], null);
			ft.commit();
			ft.begin();
			for (int i = 0; i < pobjId.length; i++) {
				pobjVal[i].setValue(NEWVALUE);
				ft.write(POBJDIR, pobjId[i], pobjVal[i], null);
			}
			ft.rollback();
			for (int i = 0; i < pobjId.length; i++)
				pobjVal[i].setValue(NULLVALUE);
			ft.begin();
			for (int i = 0; i < pobjId.length; i++)
				ft.read(POBJDIR, pobjId[i], pobjVal[i], null);
			ft.commit();
			fm.releaseTxContext(ft);
			for (int i = 0; i < pobjId.length; i++)
				if (pobjVal[i].getValue() != INITIALVALUE)
					fail("Modification rolled back: wrong value!");
		} catch (FosException fe) {
			if (fe.getNestedException() != null)
				fe.getNestedException().printStackTrace();
			fe.printStackTrace();
			fail();
		}
	}

	/**
	 * Tests the scan of the extension of a directory with no object.
	 */
	public void testScanNoObj() {
		try {
			FosTransaction ft = fm.createTxContext();
			ft.begin();
			ft.write(POBJDIR, pobjId[0], pobjVal[0], null);
			ft.commit();
			ft.begin();
			ft.delete(POBJDIR, pobjId[0]);
			ft.commit();
			ft.begin();
			Iterator it = ft.scan(POBJDIR);
			ft.commit();
			fm.releaseTxContext(ft);
			if (it.hasNext())
				fail("Extension is not composed as required.");
		} catch (FosException fe) {
			if (fe.getNestedException() != null)
				fe.getNestedException().printStackTrace();
			fe.printStackTrace();
			fail();
		}
	}

	/**
	 * Tests the scan of the extension of a directory with many objects.
	 */
	public void testScanManyObj() {
		try {
			FosTransaction ft = fm.createTxContext();
			ft.begin();
			for (int i = 0; i < pobjId.length; i++)
				ft.write(POBJDIR, pobjId[i], pobjVal[i], null);
			ft.commit();
			ft.begin();
			Iterator it = ft.scan(POBJDIR);
			ft.commit();
			fm.releaseTxContext(ft);
			boolean ex = true;
			int no = 0;
			while (it.hasNext()) {
				String pobj = (String) it.next();
				boolean found = false;
				for (int i = 0; i < pobjId.length; i++)
					if (pobjId[i].equals(pobj)) {
						found = true;
						no++;
					}
				ex &= found;
			}
			if ((!ex) || (no != pobjId.length))
				fail("Extension is not composed as required.");
		} catch (FosException fe) {
			if (fe.getNestedException() != null)
				fe.getNestedException().printStackTrace();
			fe.printStackTrace();
			fail();
		}
	}

	/**
	 * Tests if the TX context correctly pass through when the FosStructure is
	 * activated.
	 */
	public void testConnPassThrough() {
		try {
			FosTransaction ft = fm.createTxContext();
			ft.begin();
			tx = null;
			ft.write(POBJDIR, pobjId[0], pobjVal[0], null);
			if (tx != ft)
				fail();
			ft.commit();
			ft.begin();
			tx = null;
			ft.read(POBJDIR, pobjId[0], pobjVal[0], null);
			if (tx != ft)
				fail();
			ft.commit();
			fm.releaseTxContext(ft);
		} catch (FosException fe) {
			if (fe.getNestedException() != null)
				fe.getNestedException().printStackTrace();
			fe.printStackTrace();
			fail();
		}
	}

	// For managing the TestCase as a standalone Test Suite.

	/**
	 * Constructs the suite containing all test cases defined within
	 * TestTxOps.
	 */
	public static Test suite() {
		return new TestSuite(TestTxOps.class);
	}

	/**
	 * For running the TestTxOps suite standalone.
	 */
	public static void main(String args[]) {
		TestRunner.run(suite());
	}

	/**
	 * Constructs a TestTxOps.
	 */
	public TestTxOps(String tn) throws Exception {
		super(tn);
		Properties prop = new Properties();
		prop.load(ClassLoader.getSystemResourceAsStream("owfos.properties"));
		dbDir = prop.getProperty("dbdir", null);
		if (dbDir == null) {
			throw new Exception(
				"Cannot find the 'dbdir' property");
		}
	}

	/**
	 * The class that implements the I/O for a serialized  "int" Object File.
	 */
	class IntPObj implements FosStructure {
		/**
		 * The value of the integer persistent object.
		 */
		private int value = INITIALVALUE;

		/**
		 * Reads the content of a persistent object from the file represented by
		 * the given ObjectInputStream.
		 * @param ois	The ObjectInputStream representing the file from which to
		 *				read.
		 * @param conn	The connection used to access the File Object Store.
		 */
		public void readFile(ObjectInputStream ois, FosAccess conn, Object ctxt) throws Exception {
			value = ois.readInt();
			tx = conn;
		}

		/**
		 * Writes the content of a persistent object to the file represented by
		 * the given ObjectOutputStream.
		 * @param oos	The ObjectOutputStream representing the file to which to
		 *				write.
		 * @param conn	The connection used to access the File Object Store.
		 */
		public void writeFile(ObjectOutputStream oos, FosAccess conn, Object ctxt) throws Exception {
			oos.writeInt(value);
			tx = conn;
		}

		/**
		 * Returns the value of this integer persistent object.
		 */
		int getValue() {
			return value;
		}

		/**
		 * Assigns the value of this integer persistent object.
		 */
		void setValue(int val) {
			value = val;
		}
	}
}