/**
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 */

package org.objectweb.perseus.fos;

import junit.framework.Test;
import junit.framework.TestSuite;
import junit.textui.TestRunner;
import org.objectweb.perseus.LogBoot;
import org.objectweb.perseus.fos.api.FosException;
import org.objectweb.perseus.fos.api.FosTransaction;
import org.objectweb.perseus.fos.lib.FosTxContextFactory;

import java.util.Iterator;

/**
 * Tests standard transactions that correctly executes (correct begin/commit
 * sequences).
 */
public class TestTxOpsUniqMan extends TestTxOps {
	// IMPLEMENTATION OF METHODS OF THE TestCase CLASS

	/**
	 * Initialisation sequence executed before each test case.
	 */
	protected void setUp() {
	}

	/**
	 * Termination sequence executed after each test case.
	 */
	protected void tearDown() {
	}

	/**
	 * Tests rollback of a simple creation.
	 */
	public void testRollbackCreation() {
		try {
			FosTransaction ft = fm.createTxContext();
			ft.begin();
			if (ft.exist(POBJDIR, pobjId[0]))
				ft.delete(POBJDIR, pobjId[0]);
			ft.commit();
			ft.begin();
			ft.write(POBJDIR, pobjId[0], pobjVal[0], null);
			ft.rollback();
			ft.begin();
			boolean ex = ft.exist(POBJDIR, pobjId[0]);
			ft.commit();
			fm.releaseTxContext(ft);
			if (ex)
				fail("Object should not exist: creation rolled back!");
		} catch (FosException fe) {
			if (fe.getNestedException() != null)
				fe.getNestedException().printStackTrace();
			fe.printStackTrace();
			fail();
		}
	}

	/**
	 * Tests rollback of multiple creations.
	 */
	public void testRollbackCreations() {
		try {
			FosTransaction ft = fm.createTxContext();
			ft.begin();
			for (int i = 0; i < pobjId.length; i++)
				if (ft.exist(POBJDIR, pobjId[i]))
					ft.delete(POBJDIR, pobjId[i]);
			ft.commit();
			ft.begin();
			for (int i = 0; i < pobjId.length; i++)
				ft.write(POBJDIR, pobjId[i], pobjVal[i], null);
			ft.rollback();
			ft.begin();
			boolean ex = false;
			for (int i = 0; i < pobjId.length; i++)
				ex |= ft.exist(POBJDIR, pobjId[i]);
			ft.commit();
			fm.releaseTxContext(ft);
			if (ex)
				fail("Objects should not exist: creations rolled back!");
		} catch (FosException fe) {
			if (fe.getNestedException() != null)
				fe.getNestedException().printStackTrace();
			fe.printStackTrace();
			fail();
		}
	}

	/**
	 * Tests a simple exist test where the object does not exist.
	 */
	public void testExistNoObj() {
		try {
			FosTransaction ft = fm.createTxContext();
			ft.begin();
			try {
				ft.delete(POBJDIR, pobjId[0]);
			} catch (FosException fe) {
			}
			ft.commit();
			ft.begin();
			boolean ex = ft.exist(POBJDIR, pobjId[0]);
			ft.commit();
			fm.releaseTxContext(ft);
			if (ex)
				fail("Objects should not exist: no object!");
		} catch (FosException fe) {
			if (fe.getNestedException() != null)
				fe.getNestedException().printStackTrace();
			fe.printStackTrace();
			fail();
		}
	}

	/**
	 * Tests the scan of the extension of a directory with no object.
	 */
	public void testScanNoObj() {
		try {
			FosTransaction ft = fm.createTxContext();
			ft.begin();
			for (int i = 0; i < pobjId.length; i++)
				if (ft.exist(POBJDIR, pobjId[i]))
					ft.delete(POBJDIR, pobjId[i]);
			ft.commit();
			ft.begin();
			Iterator it = ft.scan(POBJDIR);
			ft.commit();
			fm.releaseTxContext(ft);
			if (it.hasNext())
				fail("Extension is not composed as required.");
		} catch (FosException fe) {
			if (fe.getNestedException() != null)
				fe.getNestedException().printStackTrace();
			fe.printStackTrace();
			fail();
		}
	}

	// For managing the TestCase as a standalone Test Suite.

	/**
	 * Constructs the suite containing all test cases defined within
	 * TestTxOpsUniqMan.
	 */
	public static Test suite() {
		return new TestSuite(TestTxOpsUniqMan.class);
	}

	/**
	 * For running the TestTxOpsUniqMan suite standalone.
	 */
	public static void main(String args[]) {
		TestRunner.run(suite());
	}

	/**
	 * Constructs a TestTxOpsUniqMan.
	 */
	public TestTxOpsUniqMan(String tn) throws Exception {
		super(tn);
		try {
			fm = new FosTxContextFactory();
			fm.setDbDir(dbDir);
			fm.setMonologFactory(LogBoot.getLoggerFactory());
			fm.initialize();
			fm.start();
		} catch (FosException fe) {
			System.out.println("Error during setUp - 1");
			if (fe.getNestedException() != null)
				fe.getNestedException().printStackTrace();
			fe.printStackTrace();
			System.exit(0);
		} catch (Exception e) {
			System.out.println("Error during setUp - 2");
			e.printStackTrace();
			System.exit(0);
		}
	}
}